import numpy as np
from .utils import *
import sympy
import scipy


# compute the refitted function, with all parameters at their best fit values, or have one of them shifted +/-1sigma
def func_evaluate(
    func_candidate,
    x,
    dim,
    param_shifted = None,
    sigma_pm = None,
    evaluate_pysr = False
):
    '''
    Compute the function values corresponding to the input array x.
    
    Arguments
    ---------
    func_candidate (pd.dataframe):
        A particular function candidate (single row of the func_candidates dataframe).
        
    x0 (np.ndarray):
        Numpy array of the independent variable.
        
    param_shifted (str):
        Set to 'a3' to evaluate the function values shifting the 'a3' while keeping the rest parameters unshifted.
        If None, evaluate the function values with all parameters at their best-fit values.
        
    sigma_pm (str):
        Set to '+' ('-1') for +1 (-1) sigma of param_shifted.
        If None, evaluate the function values with all parameters at their best-fit values.
        
    evaluate_pysr (bool):
        True: evaluate with the original function from PySR.
        False: evaluate with the refitted function from LMFIT.
    
    
    Returns
    -------
    y_pred (np.ndarray):
        The predicted dependent variable y for the function candidate.
    '''
    
    if param_shifted is None:
        # Substitute with the parameters from LMFIT (second fit) or from PySR (first fit).
        if evaluate_pysr == False:
            func = re.sub(r'\b(a\d+)\b',
                        r"func_candidate['Parameters: (best-fit, +1, -1)']['\1'][0]",
                        func_candidate['Parameterized equation, unscaled'])
                        
        else:
            func = re.sub(r'\b(a\d+)\b',
                        r"func_candidate['Parameterization']['\1']",
                        func_candidate['Parameterized equation, unscaled'])
                        
    else:
        # First substitute with all best-fit parameters.
        func = re.sub(r'\b(a\d+)\b',
                      r"func_candidate['Parameters: (best-fit, +1, -1)']['\1'][0]",
                      func_candidate['Parameterized equation, unscaled'])
                      
        # Then replace the param_shifted with its +/-1 sigma value.
        if sigma_pm == '+':
            func = re.sub(r"func_candidate\['Parameters: \(best-fit, \+1, -1\)'\]\['" + re.escape(param_shifted) + r"'\]\[0\]",
                          r"(func_candidate['Parameters: (best-fit, +1, -1)']['" + param_shifted + r"'][0] + func_candidate['Parameters: (best-fit, +1, -1)']['" + param_shifted + r"'][1])",
                          func)
                          
        elif sigma_pm == '-':
            func = re.sub(r"func_candidate\['Parameters: \(best-fit, \+1, -1\)'\]\['" + re.escape(param_shifted) + r"'\]\[0\]",
                          r"(func_candidate['Parameters: (best-fit, +1, -1)']['" + param_shifted + r"'][0] + func_candidate['Parameters: (best-fit, +1, -1)']['" + param_shifted + r"'][2])",
                          func)
    
    if dim > 1:
        for i in range(dim):
            globals()[f'x{i}'] = np.reshape(x[:, i], (-1, 1))
            
    else:
        x0 = x
        
    if re.findall(r'x\d+', func_candidate['Parameterized equation, unscaled']):
        return eval(func)
        
    else:
        # For function not depending on x.
        return np.full((x.shape[0], 1), eval(func))
        
        
def func_sampling_1d(
    func_str,
    param,
    cov,
    x,
    x_finer,
    n_samples
):
    '''
    Perform Monte Carlo sampling of the given function by
    sampling the parameters from their best fit values
    and covariance matrix.
    
    Arguments
    ---------
    func_str (str):
        Parameterized function in string format.
        
    param (dict):
        Parameter dictionary storing the best-fit and up/down unc.
        
    cov (dict):
        Covariance matrix.
        
    x (np.ndarray):
        Numpy array of the independent variable.
        
    n_samples (int):
        Number of samples.
    '''
    x0 = sympy.symbols('x0')
    param_symbols = {p: sympy.symbols(p) for p in param.keys()}
    
    func_expr = eval(func_str,
                     {**param_symbols,
                        'x0': x0,
                        'sin': sympy.sin, 'cos': sympy.cos, 'tan': sympy.tan,
                        'exp': sympy.exp, 'sinh': sympy.sinh, 'cosh': sympy.cosh, 'tanh': sympy.tanh,
                        'log': sympy.log,
                        'gauss': lambda arg: sympy.exp(-arg**2),
                        'sigmoid': lambda arg: 1./(1.+sympy.exp(-arg)),}
                     )
                     
    # Substitute parameter when it was held fixed in the fit.
    varied_params = []
    varied_params_values = []
    for p, (best_fit, up, down) in param.items():
        if up == 0 and down == 0:
            # Treat as constant parameter and substitute directly.
            func_expr = func_expr.subs(param_symbols[p], best_fit)
            
        else:
            # Add variable parameter to parameter list and values.
            varied_params.append(param_symbols[p])
            varied_params_values.append(best_fit)
            
    # Convert function to a numerical function.
    func = sympy.lambdify((x0, *varied_params), func_expr, 'numpy')
    
    # Fill the covariance matrix for varied parameters only.
    cov_matrix = np.zeros((len(varied_params), len(varied_params)))
    
    for i, p1 in enumerate(varied_params):
        for j, p2 in enumerate(varied_params):
            key = f"{p1}, {p2}"
            
            if i == j:
                # Variance (diagonal)
                if cov:
                    cov_matrix[i, j] = cov.get(f"{p1}, {p1}", 0)
                else:
                    cov_matrix[i, j] = param[f"{p1}"][1]**2
                
            else:
                # Covariance (off-diagonal)
                cov_matrix[i, j] = cov.get(key, 0)
                
    # Monte Carlo Sampling.
    samples = scipy.stats.multivariate_normal.rvs(mean = varied_params_values,
                                                  cov = cov_matrix,
                                                  size = n_samples
                                                  )
                                                  
    if len(samples.shape) == 1:
        samples = samples.reshape(-1, 1)
    
    # Evaluate the function for each parameter sample on the original and finer grid.
    func_samples = np.array([func(x, *sample) for sample in samples])
    func_samples_finer = np.array([func(x_finer, *sample) for sample in samples])
    
    # Calculate the mean and +/-1 sigma at each point in the grids
    mean_func = np.mean(func_samples, axis=0)
    lower_2sigma = np.percentile(func_samples, 2.5, axis=0)
    lower_1sigma = np.percentile(func_samples, 16, axis=0)
    upper_1sigma = np.percentile(func_samples, 84, axis=0)
    upper_2sigma = np.percentile(func_samples, 97.5, axis=0)
    
    mean_func_finer = np.mean(func_samples_finer, axis=0)
    lower_2sigma_finer = np.percentile(func_samples_finer, 2.5, axis=0)
    lower_1sigma_finer = np.percentile(func_samples_finer, 16, axis=0)
    upper_1sigma_finer = np.percentile(func_samples_finer, 84, axis=0)
    upper_2sigma_finer = np.percentile(func_samples_finer, 97.5, axis=0)
    
    func_bands = (mean_func,
                  lower_2sigma,
                  lower_1sigma,
                  upper_1sigma,
                  upper_2sigma
                  )
                  
    func_bands_finer = (mean_func_finer,
                        lower_2sigma_finer,
                        lower_1sigma_finer,
                        upper_1sigma_finer,
                        upper_2sigma_finer
                        )
    
    return func_bands, func_bands_finer
    
        
def add_gof(
    func_candidates,
    x,
    y,
    y_up,
    y_down,
    dim
):
    '''
    Compute goodness-of-fit metrics for all function candidates at once.
    
    Arguments
    ---------
    func_candidates (pd.dataframe):
        The full dataframe containing all function candidates.
        
    x (np.ndarray):
        Numpy array of the independent variable.
        
    y (np.ndarray):
        Numpy array of the dependent variable (input at central).
        
    y_up (np.ndarray):
        Numpy array of the dependent variable (input at +1 sigma).

    y_down (np.ndarray):
        Numpy array of the dependent variable (input at -1 sigma).
        
    dim (np.int):
        Dimension of the input data.
    
    
    Returns
    -------
    func_candidates (pd.dataframe):
        Add new columns to the dataframe:
            RMSE (before refit): root-mean-square-error of the PySR function before refit with LMFIT,
            RMSE: after refit with LMFIT,
            R2: coefficient of determination,
            NDF: number of degrees of freedom,
            Chi2 (before refit): Chi2 of the PySR function before refit with LMFIT,
            Chi2: after refit with LMFIT,
            Chi2/NDF (before refit): similar as above,
            Chi2/NDF: similar as above.
    '''
    # Functions after ROF.
    rmse_values = []
    r2_values = []
    chi2_values = []
    ndf_values = []
    chi2_ndf_values = []
    p_values = []
    
    # Functions before ROF.
    chi2_values_pysr = []
    chi2_ndf_values_pysr = []
    p_values_pysr = []
    rmse_values_pysr = []

    for i in range(len(func_candidates)):
        func_candidate = func_candidates.iloc[i]
        
        # Function evaluated with re-fitted parameters from LMFIT.
        y_pred = func_evaluate(func_candidate, x, dim)
        
        # Function evaluated with original parameters from PySR.
        y_pred_pysr = func_evaluate(func_candidate, x, dim, evaluate_pysr = True)
        
        rmse = np.sqrt(np.sum((y - y_pred)**2) / y.size)
        r2 = 1 - np.sum((y - y_pred)**2) / np.sum((y - np.mean(y))**2)
            
        rmse_values.append(round_a_number(rmse, 4))
        r2_values.append(round_a_number(r2, 4))
        
        # The uncertainties of input y are required to compute Chi2.
        if y_up is not None and y_down is not None:
            residual = y_pred - y
            residual_pysr = y_pred_pysr - y
            
            # For each bin, take up/down uncertainty if the residual error is +/-ve.
            # If either uncertainty input is 0, take the other one.
            y_unc = np.where(residual > 0,
                             np.where(y_up != 0, y_up, y_down),
                             np.where(y_down != 0, y_down, y_up))
                             
            y_unc_pysr = np.where(residual_pysr > 0,
                                  np.where(y_up != 0, y_up, y_down),
                                  np.where(y_down != 0, y_down, y_up))
                                  
            chi2 = np.sum(residual**2 / y_unc**2)
            chi2_pysr = np.sum(residual_pysr**2 / y_unc_pysr**2)
            
            # NDF = number of independent data point - number of varying parameters in the function.
            num_free_param = 0
            if len(func_candidate['Parameters: (best-fit, +1, -1)']) > 0:
                for j in range(len(func_candidate['Parameters: (best-fit, +1, -1)'])):
                    if func_candidate['Parameters: (best-fit, +1, -1)'][f'a{j+1}'][1] > 0:
                        num_free_param += 1
                        
            if y.size - num_free_param > 0:
                ndf = y.size - num_free_param
                
            else:
                ndf = -1
                
            p_value = scipy.stats.chi2.sf(chi2, ndf)
            p_value_pysr = scipy.stats.chi2.sf(chi2_pysr, ndf)
                
            ndf_values.append(ndf)
            chi2_values.append(round_a_number(chi2, 4))
            chi2_ndf_values.append(round_a_number(chi2/ndf, 4))
            chi2_values_pysr.append(round_a_number(chi2_pysr, 4))
            chi2_ndf_values_pysr.append(round_a_number(chi2_pysr/ndf, 4))
            p_values.append(round_a_number(p_value, 4))
            p_values_pysr.append(round_a_number(p_value_pysr, 4))
            
        else:
            rmse_pysr = np.sqrt(np.sum((y - y_pred_pysr)**2) / y.size)
            rmse_values_pysr.append(round_a_number(rmse_pysr, 4))
        
    func_candidates['RMSE'] = rmse_values
    func_candidates['R2'] = r2_values
    
    if y_up is not None and y_down is not None:
        func_candidates['NDF'] = ndf_values
        
        func_candidates['Chi2 (before ROF)'] = chi2_values_pysr
        func_candidates['Chi2'] = chi2_values
        
        func_candidates['Chi2/NDF (before ROF)'] = chi2_ndf_values_pysr
        func_candidates['Chi2/NDF'] = chi2_ndf_values
        
        func_candidates['p-value (before ROF)'] = p_values_pysr
        func_candidates['p-value'] = p_values
        
    else:
        func_candidates['RMSE (before ROF)'] = rmse_values_pysr
    
    return func_candidates
    



import numpy as np
from .utils import round_a_number
    

def dataset_formatting(
    x,
    y,
    y_up,
    y_down,
    fit_y_unc
):

    dim = 0
    
    # Input dataset is in python lists.
    if isinstance(x, list):
        # dataset[0] is the independent variable x that takes the form:
        # [1, 2, 3,...] of 1D;
        # [[1, 1], [1, 2], [1, 3],...] for 2D;
        # [[1, 1, 1], [1, 1, 2], [1, 1, 3],...] for 3D etc.
        if isinstance(x[0], list):
            dim = len(x[0])
            
        else:
            dim = 1
            
    elif isinstance(x, np.ndarray):
        # dataset[0] is the independent variable in np.ndarray with shape (n, dim).
        dim = np.shape(x)[1]
        
    else:
        raise TypeError('Input x should be either lists or numpy arrays.')
            
    # Get x and y.
    x = np.reshape(np.array(x), (-1, dim))
    y = np.reshape(np.array(y), (-1, 1))
    
    assert(x.shape[0] == y.shape[0])
        
    # Get uncertainties for y.
    if y_up is not None and y_down is not None:
        #if isinstance(dataset[2], list) and isinstance(dataset[3], list):
        #    pass
        #elif isinstance(dataset[2], np.ndarray) and isinstance(dataset[3], np.ndarray):
        #    pass
        #else:
        #    raise TypeError('Input dataset should be either python lists or numpy arrays.')
        
        y_up = np.reshape(np.array(y_up), (-1, 1))
        y_down = np.reshape(np.array(y_down), (-1, 1))
        
        assert(x.shape[0] == y.shape[0] == y_up.shape[0] == y_down.shape[0])

        if fit_y_unc:
            # Do not fit with y uncert. if any of these is zero,
            # since loss = (y_pred - y_true)^2/y_unc^2 would become undefined.
            if np.any(np.logical_and(y_up == 0, y_down == 0)):
                y_up = None
                y_down = None
                fit_y_unc = False
                
        else:
            y_up = None
            y_down = None
            fit_y_unc = False
            
    else:
        y_up = None
        y_down = None
        fit_y_unc = False
        
    return x, y, y_up, y_down, fit_y_unc, dim
        

def histogram_scale(
    x,
    y,
    y_up,
    y_down,
    x_min = 0,
    x_max = 1,
    scale_y_by = None
):
    '''
    Pre-process the input data by scaling both x and y:
        1) scale x to [x_min, x_max],
        2) scale y s.t. the normalization = norm (computed with np.trapz).
    This is to stablize the fits and avoid numerical overflow etc.
    
    Arguments
    ---------
    x (np.ndarray):
        Numpy array of the independent variable.
        
    y (np.ndarray):
        Numpy array of the dependent variable (input at central).
        
    y_up (np.ndarray):
        Numpy array of the dependent variable (input at +1 sigma).

    y_down (np.ndarray):
        Numpy array of the dependent variable (input at -1 sigma).
        
    x_min (np.float):
        Minimum of x after scaling.
    
    x_max (np.float):
        Maximum of x after scaling.
    
    norm (np.float):
        Normalization of (x, y) after scaling.
    
    
    Returns
    -------
    X (np.ndarray):
        Scaled x with the range [x_min, x_max].
    
    Y (np.ndarray):
        Scaled y with normalization.
    
    Y_up (np.ndarray):
        Scaled y_up by the same norm as applied to y.
    
    Y_down (np.ndarray):
        Scaled y_down by the same norm as applied to y.
    
    y_scale (np.float):
        The scale number multiplied to y's,
        it will be used to unscale the functions after fitting.
    '''
    
    # Scale the x axis to a range from x_min to x_max.
    X = (x - np.min(x, axis = 0)) * (x_max - x_min) / (np.max(x, axis = 0) - np.min(x, axis = 0)) + x_min
    
    # Scale y.
    if scale_y_by == 'max':
        y_scale = 1 / np.abs(np.max(y))
        
    elif scale_y_by == 'mean':
        y_scale = 1 / np.abs(np.mean(y))
        
    elif scale_y_by == 'l2':
        y_scale = 1 / np.linalg.norm(y)
        
    else:
        y_scale = 1
        
    Y = y * y_scale
    
    if y_up is not None and y_down is not None:
        Y_up = y_up * y_scale
        Y_down = y_down * y_scale
        
    else:
        Y_up = None
        Y_down = None
    
    return X, Y, Y_up, Y_down, y_scale
    

def functions_unscale(
    func_candidates,
    x,
    X,
    y_scale,
    input_scale,
    dim
):
    '''
    The PySR/LMFIT fits were done on the scaled data,
    these fitted functions need to be unscaled to described the original data.
    
    Arguments
    ---------
    func_candidates (pd.dataframe):
        Dataframe containing all function candidates after fits.
        
    x (np.ndarray):
        The independent variable before scaling.
        
    X (np.ndarray):
        The independent variable after scaling.
    
    y_scale (np.float):
        The scale number multiplied to y's,
        it will be used to unscale the functions after fitting.
        
    input_rescale (bool):
        True: scaling enabled.
        False: scaling not enabled, so just return the same functions after fits.
    
    
    Returns
    -------
    func_candidates (pd.dataframe):
        Add a new column for the unscaled parameterized functions.
    '''
    
    # Scaled.
    X_min = np.min(X, axis = 0)
    X_max = np.max(X, axis = 0)
    
    # Original.
    x_min = np.min(x, axis = 0)
    x_max = np.max(x, axis = 0)
    
    X_range = X_max - X_min
    x_range = x_max - x_min
    
    func_unscaled = []
    
    for i in range(len(func_candidates)):
        func = func_candidates['Parameterized equation'][i]
        
        if input_scale == True:
            # Scale X back to the original range:
            # substitute X by (x - x_min) * (x_range_scaled / x_range_original) + X_min.
            for j in range(dim):
                var = f'x{j}'
                if X_min[j] == 0:
                    func = func.replace(var, '(({0} - {1}) * {2})'.format(var, x_min[j], round_a_number(X_range[j]/x_range[j], 6)))
                    
                else:
                    func = func.replace(var, '(({0} - {1}) * {3} + {2})'.format(var, x_min[j], X_min[j], round_a_number(X_range[j]/x_range[j], 6)))
                    
            # Scale Y back to the original scale by multiplying by 1/y_scale.
            func_unscaled.append('{0}*({1})'.format(round_a_number(1/y_scale, 6), func))
            
        else:
            func_unscaled.append(func)
            
    func_candidates['Parameterized equation, unscaled'] = func_unscaled
    
    return func_candidates
    





import pandas as pd
import sympy
import os
import re
from pysr import PySRRegressor
from glob import glob
from .math_defs import *

pd.set_option('display.max_colwidth', None)
pd.options.mode.chained_assignment = None
np.seterr(divide='ignore', invalid='ignore')


def round_a_number(number, sig_fig = 6):
    '''
    Round a numpy float number to a certain significant figure.
    '''
        
    try:
        return np.round(number, sig_fig - 1 - int(np.floor(log10(np.abs(number)))))
        
    except:
        return number


def round_numbers_in_sympy_expr(sympy_expr, sig_fig = 3):
    '''
    Round all numbers in a sympy expression to a certain significant figure.
    '''
    
    def rounding(x, sig = sig_fig):
        if x.is_Number and not x.is_Integer:
            return x.round(sig - 1 - int(sympy.floor(sympy.log(abs(x), 10))))
            
        else:
            return x
    
    rounded_expr = sympy_expr.replace(lambda x: x.is_Number,
                                      lambda x: rounding(x, sig_fig))
                                      
    return rounded_expr


def simplify_pkl(pysr_pkl, x):
    '''
    Extract the function candidates from the PySR output file (.pkl) and create a dataframe for them.
    
    Arguments
    ---------
    pysr_pkl (str):
        PySR output file.
    
    x (np.ndarray):
        Independent variable (scaled).
    
    
    Returns
    -------
    A dataframe containing the function candidates, complexity, and loss from the PySR fit.
    '''
    
    model = PySRRegressor.from_file(pysr_pkl)
    
    equation = []
    complexity = []
    loss = []
    
    for i in range(len(model.equations_['equation'])):
        # Simplify the pysr functions algebraically (e.g., (x0 + x0) -> 2 * x0).
        with sympy.evaluate(True):
            equ = model.sympy(i)
        
        # Round the float numbers to a certain significant figure.
        equ = round_numbers_in_sympy_expr(equ)
        
        # Store functions in strings for later manipulation.
        # Also replace custom sympy operators to those that can be treated with numpy.
        equ = str(equ).replace('^', '**').replace('Piecewise(', 'piecewise').replace(' > 0), (0.0, True))', ')')
        
        equation.append(equ)
        
        complexity.append(model.equations_['complexity'][i])
        
        loss.append(round_a_number(model.equations_['loss'][i] * len(x), sig_fig = 3))
        
    df = pd.DataFrame({'PySR equation': equation, 'Complexity': complexity, 'PySR loss': loss})
    
    return df




import lmfit
import importlib
import os
from lmfit import Parameters, Minimizer, fit_report, conf_interval
from itertools import combinations

from .utils import *
from .evaluate import *
from .plotting import *
from .math_defs import *
from .processing import *

import warnings
warnings.filterwarnings("ignore")


class SymbolFit:
    '''
    SymbolFit class
    
    
    Parameters
    ----------
    x : list | ndarray
        Independent variable x, or bin center values for histogram data.
        If provided as a python list,
        e.g., [1, 2, 3,...] for 1D,
        [[1, 1], [1, 2], [1, 3],...] for 2D,
        [[1, 1, 1], [1, 1, 2], [1, 1, 3],...] for 3D etc.
        If provided as ndarray, then shape is (num_examples, dim).
        
    y : list | ndarray
        Dependent variable y, or bin content values for histogram data.
        Shape is (num_examples, 1).
        
    y_up : list | ndarray
        Upper one standard deviation of y (+1 sigma).
        It should be the deviation value and non-negative.
        Shape is (num_examples, 1).
        
    y_down : list | ndarray
        Lower one standard deviation of y (-1 sigma).
        It should be the deviation value and non-negative.
        Shape is (num_examples, 1).
    
    pysr_config : pysr.PySRRegressor class
        Configuration for PySR training,
        see https://github.com/MilesCranmer/PySR.
        The configuration can be stored in a python file like pysr_config.py:
            from pysr import PySRRegressor
            pysr_config = PySRRegressor(...)
        and source from there:
            pysr_config = importlib.import_module('directory.pysr_config').pysr_config
            model = SymbolFit(..., pysr_config = pysr_config,...)
            
    max_complexity : int
        Maximum complexity of expression tree.
        Overwrite the maxsize parameter in PySRRegressor() if provided.
            
    input_rescale : bool
        Rescale x to the range of (0, 1) for fitting,
        which could avoid fit instability or overflow.
        Fitted functions will be unscaled at outputs.
        
    scale_y_by : str
        Normalize y for fitting:
        divide y by its 'max', 'mean', 'l2', or None.
        Fitted functions will be unscaled at outputs.
        Applicable when input_rescale is True.
        
    max_stderr : float (%)
        During refitting with LMFIT, fit is considered failed when
        any of the parameters has an uncertainty larger than max_stderr,
        then retry the fit by decreasing the ndf,
        keeping some parameters fixed to their initial values.
        It is to avoid bad fits when any parameters get unrealistically large.
        Note: setting max_uncertainty to O(10) suffices in most cases.
        
    fit_y_unc : bool
        Consider y_up and y_down in the fits,
        i.e., take chi2 loss = (y_pred - y_true)^2 / y_unc^2.
        Here, for each bin, y_unc is taken as y_up when (fit - y) > 0,
        and taken as y_down when (fit - y) < 0.
    
    random_seed : int
        Overwrite pysr_config:
            pysr_model.set_params(procs = 0,
                                  multithreading = False,
                                  random_state = self.random_seed,
                                  deterministic = True)
        for reproducing the same batch of candidate functions.
        This will force to run PySR in single thread, so slower.
        
    loss_weights : list | ndarray
        Scale loss by (y_model - y_label)^2 * loss_weights in fits.
        Will overwrite (y_model - y_label)^2 / y_unc^2 if provided.
        Shape is (num_examples, 1).
    '''
    
    def __init__(
        self,
        x = None,
        y = None,
        y_up = None,
        y_down = None,
        pysr_config = None,
        max_complexity = None,
        input_rescale = True,
        scale_y_by = None,
        max_stderr = 40,
        fit_y_unc = True,
        random_seed = None,
        loss_weights = None,
        func_candidates = pd.DataFrame()
    ):
        self.x = x
        self.y = y
        self.y_up = y_up
        self.y_down = y_down
        self.pysr_config = pysr_config
        self.max_complexity = max_complexity
        self.input_rescale = input_rescale
        self.scale_y_by = scale_y_by
        self.max_stderr = max_stderr
        self.fit_y_unc = fit_y_unc
        self.random_seed = random_seed
        self.loss_weights = loss_weights
        self.func_candidates = func_candidates
        
        
    def fit(
        self
    ):
        '''
        Performs a search for functional forms with PySR.
        Parameterizes constants in all functions.
        Creates a loop of re-optimization fit (ROF) to improve
        the constants and provide uncertainty estimation.
        '''
        
        x = self.x
        y = self.y
        y_up = self.y_up
        y_down = self.y_down
        pysr_model = self.pysr_config
        max_complexity = self.max_complexity
        input_rescale = self.input_rescale
        scale_y_by = self.scale_y_by
        max_stderr = self.max_stderr
        fit_y_unc = self.fit_y_unc
        random_seed = self.random_seed
        loss_weights = self.loss_weights
        
        
        x, y, y_up, y_down, fit_y_unc, dim = dataset_formatting(x = x,
                                                                y = y,
                                                                y_up = y_up,
                                                                y_down = y_down,
                                                                fit_y_unc = fit_y_unc
                                                                )
        self.fit_y_unc = fit_y_unc
            
            
        # Rescale the input data before training to prevent overflows in the fits,
        # since say exp(x) would cause problem if x is large.
        # E.g., x -> [0,1] and norm -> 1.
        if self.input_rescale:
            X, Y, Y_up, Y_down, y_scale = histogram_scale(x = x,
                                                          y = y,
                                                          y_up = y_up,
                                                          y_down = y_down,
                                                          x_min = 0,
                                                          x_max = 1,
                                                          scale_y_by = scale_y_by
                                                          )
            
        else:
            X, Y, Y_up, Y_down, y_scale = x, y, y_up, y_down, 1.
            
    
        # Remove output files from previous PySR fits before starting new one below.
        pklfiles = glob('hall*')
        for f in pklfiles:
            os.remove(f)
        
        # In PySR, set weighted loss = (y_model - y_label)^2 * loss_weights.
        if loss_weights is not None:
            loss_weights = np.reshape(np.array(loss_weights), (-1, 1))
            
        elif y_up is not None and y_down is not None:
            loss_weights = np.where(Y_up != 0, Y_up, Y_down)
            loss_weights = 1 / loss_weights**2
            
        else:
            loss_weights = np.ones(y.shape)
            
        
        # Run PySR fit.
        if self.random_seed is not None:
            pysr_model.set_params(procs = 0,
                                  multithreading = False,
                                  random_state = self.random_seed,
                                  deterministic = True)
                                  
        if self.max_complexity is not None:
            pysr_model.set_params(maxsize = max_complexity)
            
            
        pysr_model.fit(X, Y, weights = loss_weights.flatten())
        
        print('\n')
        
        # Get essential info from the PySR output file (hall_.pkl),
        # and save to a df for later processing/refit.
        os.rename(glob('hall*.pkl')[0],'pysr_model_temp.pkl')
        
        func_candidates = simplify_pkl(pysr_pkl = 'pysr_model_temp.pkl',
                                       x = X
                                       )
                                       
        print('\n')
        
            
        # The constants in the fitted functions from PySR do not have uncert. estimation,
        # so we fix the functional forms, parameterize all constants, and refit them with LMFIT.
        # The first step is to parameterize the fitted functions from PySR.
        def parameterize_func_single(func_str, dim):
            '''
            Parameterizing a function like: '1.2*x0 + exp(3.4*x0)' -> 'a1*x0 + exp(a2*x0)'.
            Set the constants as initial conditions for the parameters in the second fits:
                e.g., {'a1': 1.2, 'a2': 3.4}.
            
            Arguments
            ---------
            func_str (str):
                Function in str like '1.2*x0 + exp(3.4*x0)'.
            
            
            Returns
            -------
            function (str):
                E.g., '1.2*x0 + exp(3.4*x0)' same as input
                
            parameterized function (str):
                E.g., 'a1*x0 + exp(a2*x0)'
                
            parameterization (dict):
                E.g., {'a1': 1.2, 'a2': 3.4}
            '''
            
            # Define the independent variable (x0, x1, x2,...),
            # parse the input function str into a SymPy equation.
            for i in range(dim):
                func = sympy.sympify(func_str, locals={f'x{i}': sympy.symbols(f'x{i}')})
            
            
            # Loop over all constants in the function using SymPy's tree traversal method.
            # If there are 3 constants (1.2, 3.4, 5.6), then create {1.2: 'a1', 3.4: 'a2', 5.6: 'a3'} for them.
            parameterization = {}
            variable_counter = 0
            
            for a in sympy.preorder_traversal(func):
                # Replace non-integer constants only,
                # as sometimes constants like 1 are trivial.
                if isinstance(a, sympy.Number) and not a.is_Integer:
                    variable_counter += 1
                    parameterization[a] = sympy.Symbol(f'a{variable_counter}')


            # In case there are constants belonging to more than one subtree,
            # the earlier variable_counter will be overriden,
            # creating {1.2: 'a1', 3.4: 'a3', 5.6: 'a4'} instead of {1.2: 'a1', 3.4: 'a2', 5.6: 'a3'}.
            # We don't want that since it will cause problem in later processing.
            # Here, we rename the variables with ascending subscripts {'a1', 'a2', 'a3',...}.
            def rename_ascending_variables(param_dict):
                # parameterization was created with float keys and sympy symbols as values.
                # First sort the items by the constants (keys).
                sorted_items = sorted(param_dict.items(), key = lambda x: x[0])

                # Loop over the sorted items, rename the variables,
                # and store to a new dictionary.
                param_dict_renamed = {}
                for i, (key, value) in enumerate(sorted_items, start = 1):
                    renamed_variable = f'a{i}'
                    
                    if value.name != renamed_variable:
                        value = sympy.symbols(renamed_variable)
        
                    param_dict_renamed[key] = value
    
                return param_dict_renamed
                
            parameterization = rename_ascending_variables(parameterization)

            # Replace constants by their corresponding parameters in the function.
            func_parameterized = func.subs(parameterization)
            
            return str(func), str(func_parameterized), {str(value): float(key) for key, value in parameterization.items()}
        
        
        def parameterize_func_all(func_candidates, dim):
            '''
            Parameterize all candidate functions at once.
            
            Arguments
            ---------
            func_candidates (pd.dataframe):
                Contains all PySR functions to be parameterized.
            
            
            Returns
            -------
            func_candidates (pd.dataframe):
                Add new columns for the parameterized functions
                and the parameterization that stores the initial values.
            '''
            
            # Loop over all candidate functions, parameterize them,
            # and store as new columns.
            func_param_all = []
            param_all = []
            
            for i in range(len(func_candidates)):
                _, func_param, param = parameterize_func_single(func_str = func_candidates['PySR equation'][i],
                                                                dim = dim
                                                                )
                
                func_param_all.append(func_param)
                param_all.append(param)
                
            func_candidates['Parameterized equation'] = func_param_all
            
            func_candidates['Parameterization'] = param_all
            
            func_candidates = func_candidates[[
                'PySR equation',
                'Parameterized equation',
                'Parameterization',
                'Complexity'
            ]]
            
            return func_candidates


        def get_rel_err(param):
            '''
            Get relative error of a fitted parameter from the LMFIT result.
            
            Arguments
            ---------
            param (LMFIT.Param):
                Parameter instance of LMFIT after fit.
            
            
            Returns
            -------
            value (np.float):
                Best-fit.
            
            std_err (np.float):
                Standard error.
            
            ratio (np.float):
                (standard error / best-fit value) * 100, in absolute value.
            '''
            
            # Get numbers of a fitted parameter from the LMFIT result
            numbers = re.findall(r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?", str(param))
            numbers = np.array(numbers, dtype=float)
            
            value = None
            std_err = None
            ratio = 99999
            
            # If the fit succeeds, LMFIT parameter returns something like
            # "... a1 ... [best-fit] ... [standard error]...", 3 numbers in total,
            # we need the last two to calculate the relative error.
            # Otherwise the parameter has been held fixed or sent to inf,
            # which we assign 99999 for the relative error and ignore it for later.
            if len(numbers) == 3:
                value, std_err = numbers[1], numbers[2]
                ratio = std_err / np.abs(value) * 100
                
            return value, std_err, ratio
            

        def get_covariance_correlation(fit_result):
            '''
            Get correlation of all fitted parameters from the LMFIT result.
            
            Arguments
            ---------
            fit_result:
                Fit result from LMFIT.
            
            
            Returns
            -------
            correlation (dict):
                Correlation matrix of fitted parameters {'a1, a2': 0.1, ...}.
            '''
            
            # Get the number of fitted parameters that were not held fixed.
            num_param = len(fit_result.params)
            param_varied = []
            
            for i in range(num_param):
                if 'fixed' not in str(fit_result.params[f'a{i+1}']):
                    param_varied.append(f'a{i+1}')
                    
            num_param_varied = len(param_varied)
            
            # Store diagonal and upper or lower triangular elements.
            covariance = {}
            
            # Calculate correlation as covariance(x, y) / (stderr_x * stderr_y).
            correlation = {}
            
            # Correlation only when at least one pair of parameters exist.
            if num_param_varied > 1:
                # All possible pairs.
                for i in range(num_param_varied - 1):
                    for j in range(i + 1, num_param_varied):
                        # Get a pair of parameters.
                        pair = '{0}, {1}'.format(param_varied[i], param_varied[j])
                        
                        # Get standard errors from covariance matrix.
                        _, error_i, _ = get_rel_err(param = fit_result.params[param_varied[i]])
                        _, error_j, _ = get_rel_err(param = fit_result.params[param_varied[j]])
                        
                        # Compute the correlation value.
                        corr = round_a_number(number = fit_result.covar[i][j] / (error_i * error_j),
                                              sig_fig = 4
                                              )
                        
                        # Numerical rounding causing correlation slightly > 1 or < 1.
                        if corr > 1:
                            corr = 1
                            
                        elif corr < -1:
                            corr = -1
                            
                        correlation[pair] = corr
                        
                        # Off diagonal elements of the covariance matrix.
                        covariance[pair] = fit_result.covar[i][j]
                        
                # Diagonal elements of the covariance matrix.
                for i in range(num_param_varied):
                    pair = '{0}, {1}'.format(param_varied[i], param_varied[i])
                    covariance[pair] = fit_result.covar[i][i]
                        
            return covariance, correlation
            
            
        def refit_single(func_candidate, x, y, y_up, y_down, max_stderr, dim):
            '''
            After parameterizing a PySR function, fit it with LMFIT.
            It contains a refit loop:
                First allow all parameters to vary in the fit,
                if the fit fails (no min found or any parameter returns a too high error)
                then loosen the ndf (keep some of parameters fixed in fits) and refit again.
            
            Arguments
            ---------
            func_candidate (pd.dataframe):
                Contains one candidate function (one row).
                
            x (np.ndarray):
                Independent variable of data after processing.
            
            y (np.ndarray):
                Dependent variable of data after processing.
                
            y_up (np.ndarray):
                +1 sigma for y.
            
            y_down (np.ndarray):
                -1 sigma for y.
            
            max_stderr (np.float):
                Maximum standard error (%) allowed for all parameters in a fit.
                If the fit returns any of the parameters with a stderr higher than that,
                the fit is re-tried by loosening the ndf by fixing some of the parameters.
            
            
            Returns
            -------
            result:
                Full fit result returned by the LMFIT Minimizer.
            
            covariance:
                Covariance matrix of the fitted parameters.
                
            correlation:
                Correlation matrix of the fitted parameters.
            
            ci:
                Confidence intervals for the fitted parameters
                (more robust than standard errors from Minimizer).
            '''
            
            # Define the minimization objective for LMFIT (it takes residual^2).
            def residual(params, x, data, y_up, y_down, dim):
                if dim > 1:
                    for i in range(dim):
                        globals()[f'x{i}'] = np.reshape(x[:, i], (-1, 1))
                        
                else:
                    x0 = x
                    
                # LMFIT uses Parameter object to represent a variable,
                # substitute a1, a2,... by their Parameter object defined in the input params.
                model = re.sub(r'\b(a\d+)\b', r"params['\1']", func_candidate['Parameterized equation'])
                residual = eval(model) - data
                
                # Scale residual by +/-1 sigma in the fit if they exist.
                # If the residual is +ve in a bin, scale by +1 sigma,
                # otherwise scale by -1 sigma.
                # Also replace with one another if either is zero in a bin.
                if y_up is not None and y_down is not None:
                    y_unc = np.where(residual > 0,
                                     np.where(y_up != 0, y_up, y_down),
                                     np.where(y_down != 0, y_down, y_up)
                                     )
                                     
                    residual = residual / y_unc
                    
                return residual
            
            # In principle, we want to have all parameters being varied in a fit,
            # but sometimes the fit fails due to instability or hitting infinities.
            # In these cases, we fit with less ndf by keeping some parameters fixed in next fits.
            # So start a fit with all parameters being variables.
            # If it fails or any parameters have too high relative errors (> max_stderr),
            # redo the fit by fixing one of more parameters until the fit succeeds.
            
            # Here, we create a list that contains all possible combinations of
            # varying/fixing parameters in a fit.
            # If there are two parameters in a function, then it will generate:
            # [[True, True], [True, False], [False, True], [False, False]].
            def vary_combinations(num_params):
                # Start from a list with all True's.
                vary_combo = [[True] * num_params]
                
                # Create new combinations by turning some into False's.
                for r in range(1, num_params + 1):
                    for combo in combinations(range(num_params), r):
                        temp_list = vary_combo[0].copy()
                        
                        for index in combo:
                            temp_list[index] = False
                            
                        vary_combo.append(temp_list)
                        
                return vary_combo
            
            num_params = len(func_candidate['Parameterization'])
            
            # In case the function does not have any parameter to fit, e.g. y=1 or y=exp(x).
            if num_params == 0:
                return None
            
            vary_combo = vary_combinations(num_params)
            
            
            # Loop over the possible combinations of which parameters to vary/fixed in a fit.
            for vary_trial in range(len(vary_combo)):
                print('    >>> loop of re-parameterization with less NDF for bad fits {0}/{1}...'.format(vary_trial + 1, len(vary_combo)), end='\r')
                
                params = Parameters()
                
                for i in range(num_params):
                    # Define the LMFIT Parameter for the current fit,
                    # setting which parameters to vary/fixed from the vary_combo,
                    # and their initial values from the parameterization.
                    init_value = func_candidate['Parameterization'][f'a{i+1}']
                    
                    # LMFIT Parameter class method add().
                    params.add(name = f'a{i+1}',
                               value = init_value,
                               vary = vary_combo[vary_trial][i]
                               )
                    
                try:
                    # Fit the parameters with the LMFIT minimizer.
                    mini = Minimizer(userfcn = residual,
                                     params = params,
                                     fcn_args = (x, y, y_up, y_down, dim))
                                     # ^ arguments for residual()
                                     
                    result = mini.minimize()
                    
                except Exception as e:
                    # The fit might not converge due to complex phase space,
                    # in that case we give up the current vary_combo and
                    # retry another one (decreasing ndf by fixing more parameters).
                    continue
                    
                # If the above fit converges, get the standard errors in %.
                rel_errors = []
                
                for i in range(num_params):
                    if vary_combo[vary_trial][i]:
                        _, _, ratio = get_rel_err(param = result.params[f'a{i+1}'])
                        
                        rel_errors.append(ratio)
                
                # Here, check if all relative errors are within the pre-set max_stderr,
                # if it does, then stop the refit loop and go to confidence interval calculations,
                # otherwise continue the loop (retry another vary_combo).
                # A finite/small max_stderr is to ensure bounded errors for the fitted parameters.
                # Also, we don't want a too small max_stderr, since it would lead to only
                # a small subset of parameters that float and survive the fit, and the
                # errors would be too small for a meaningful uncertainty model.
                if len(rel_errors) > 0 and all(rel_error < max_stderr for rel_error in rel_errors):
                    print('    >>> loop of re-parameterization with less NDF for bad fits {0}/{1}...\n'.format(vary_trial + 1, len(vary_combo)))
                    
                    break
                    
            # Compute the correlation for fitted parameters from the standard error estimation.
            covariance, correlation = get_covariance_correlation(fit_result = result)
            
            # Compute the confidence intervals for the fitted parameters,
            # which are more robust estimation of the uncertainties than standard errors.
            if len(rel_errors) > 1:
                try:
                    #ci = lmfit.conf_interval(minimizer = mini,
                    #                         result = result
                    #                         )
                    ci = None
                    
                except:
                    ci = None
                    
            else:
                ci = None

            return result, covariance, correlation, ci

        
        def refit_all(func_candidates, x, y, y_up, y_down, max_stderr, dim):
            '''
            Fit all parameterized candidate functions at once, using the
            refit_single(func_candidate, x, y, y_up, y_down, max_stderr) defined above.
            
            Arguments
            ---------
            func_candidates (pd.dataframe):
                Contains all candidate functions.
                
            x (np.ndarray):
                Independent variable of data after processing.
            
            y (np.ndarray):
                Dependent variable of data after processing.
                
            y_up (np.ndarray):
                +1 sigma for y.
            
            y_down (np.ndarray):
                -1 sigma for y.
            
            max_stderr (np.float):
                Maximum standard error (%) allowed for all parameters in a fit.
                If the fit returns any of the parameters with a stderr higher than that,
                the fit is re-tried by loosening the ndf by fixing some of the parameters.
            
            
            Returns
            -------
            func_candidates (pd.dataframe):
                Update the dataframe by adding new columns:
                    1) 'Parameters: (best-fit, +1, -1)',
                    2) 'Covariance',
                    3) 'Correlation',
                    4) 'Confidence interval'.
            '''
            
            def get_val_err(param_str):
                # Get relevant numbers from Parameter in string.
                numbers = re.findall(r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?", param_str)
                numbers = np.array(numbers, dtype = float)
                
                # If fit returns, there will be 3 numbers.
                # First is best-fit, then up and down unc.
                if len(numbers) == 3:
                    value, error = numbers[1], numbers[2]
                    
                else:
                    value = numbers[1]
                    error = 0
                    
                return value, error
            
            refitted_params = []
            covariances = []
            correlations = []
            confidence_intervals = []
            
            for i in range(len(func_candidates)):
                print('Re-optimizing parameterized candidate function {0}/{1}...'.format(i + 1, len(func_candidates)))
                
                # Do nothing for candidate without any parameter to fit.
                if func_candidates['Parameterization'][i] == {}:
                    refitted_params.append({})
                    
                    covariances.append({})
                    correlations.append({})
                    confidence_intervals.append(None)
                    
                else:
                    refitted_param = {}
                    
                    # Run the refit.
                    result, covariance, correlation, ci = refit_single(func_candidate = func_candidates.iloc[i],
                                                                       x = x,
                                                                       y = y,
                                                                       y_up = y_up,
                                                                       y_down = y_down,
                                                                       max_stderr = max_stderr,
                                                                       dim = dim
                                                                       )
                    
                    # Either all from standard least-square or all from ci.
                    for j in range(len(result.params)):
                        #value, error = get_val_err(str(result.params[f'a{j+1}']))
                        #refitted_param[f'a{j+1}'] = round_a_number(value, sig_fig=6), round_a_number(error, sig_fig=6)
                        try:
                            # See LMFIT confidence intervals output format.
                            central = ci[f'a{j+1}'][3][1]
                            up = ci[f'a{j+1}'][4][1] - central
                            down = ci[f'a{j+1}'][2][1] - central
                            
                            refitted_param[f'a{j+1}'] = round_a_number(central), round_a_number(up), round_a_number(down)
                            
                        except:
                            central, std_err = get_val_err(param_str = str(result.params[f'a{j+1}']))
                            
                            refitted_param[f'a{j+1}'] = round_a_number(central), round_a_number(std_err), round_a_number(-std_err)
                        
                    refitted_params.append(refitted_param)
                    covariances.append(covariance)
                    correlations.append(correlation)
                    confidence_intervals.append(ci)
                    
                    
            func_candidates['Parameters: (best-fit, +1, -1)'] = refitted_params
            func_candidates['Covariance'] = covariances
            func_candidates['Correlation'] = correlations
            func_candidates['Confidence interval'] = confidence_intervals
            
            func_candidates = func_candidates[[
                'Complexity',
                'PySR equation',
                'Parameterized equation',
                'Parameterization',
                'Parameters: (best-fit, +1, -1)',
                'Covariance',
                'Correlation'
            ]]
                    
            return func_candidates
        
        
        # Parameterize all functional forms from PySR outputs.
        func_candidates = parameterize_func_all(func_candidates = func_candidates,
                                                dim = dim
                                                )
        
        # Re-optimization loop (ROF) to improve constants and provide unc estimation.
        func_candidates = refit_all(func_candidates = func_candidates,
                                    x = X,
                                    y = Y,
                                    y_up = Y_up,
                                    y_down = Y_down,
                                    max_stderr = max_stderr,
                                    dim = dim
                                    )
        
        # Undo the input rescaling after all the fits.
        func_candidates = functions_unscale(func_candidates = func_candidates,
                                            x = x,
                                            X = X,
                                            y_scale = y_scale,
                                            input_scale = input_rescale,
                                            dim = dim
                                            )
        
        # Compute goodness-of-fit scores.
        func_candidates = add_gof(func_candidates = func_candidates,
                                  x = x,
                                  y = y,
                                  y_up = y_up,
                                  y_down = y_down,
                                  dim = dim
                                  )
        
        
        # Remove intermediate files.
        intermediate_files = glob('hall*')
        for f in intermediate_files:
            os.remove(f)
            
        os.remove('pysr_model_temp.pkl')
        
        # Update the full func_candidates dataframe containing all results.
        self.func_candidates = func_candidates
        
        
    def save_to_csv(
        self,
        output_dir = './',
    ):
        '''
        Saves the func_candidates dataframe (results) to a csv file.
        1) Full info -> candidates.csv.
        2) Reduced info -> candidates_reduced.csv.
        
        Arguments
        ---------
        output_dir : str
            Output directory.
        '''
        
        func_candidates = self.func_candidates
        
        # Define output directory to store csv and pdf files.
        if output_dir.endswith('/'):
            output_dir = output_dir
            
        else:
            output_dir = output_dir + '/'
            
        os.makedirs(output_dir) if not os.path.exists(output_dir) else None
        
        
        # Save the full func_candidates dataframe to a csv file.
        print('Saving full results >>> {}candidates.csv'.format(output_dir))
        
        func_candidates.to_csv(output_dir + 'candidates.csv')
        
        # Save the reduced version removing unnecessary info.
        print('Saving reduced results >>> {}candidates_reduced.csv'.format(output_dir))
        
        try:
            func_candidates[[
                'Parameterized equation, unscaled',
                'Parameters: (best-fit, +1, -1)',
                'Covariance',
                'Correlation',
                'RMSE',
                'R2',
                'NDF',
                'Chi2',
                'Chi2/NDF',
                'p-value'
            ]].to_csv(output_dir + 'candidates_reduced.csv')
            
        except:
            func_candidates[[
                'Parameterized equation, unscaled',
                'Parameters: (best-fit, +1, -1)',
                'Covariance',
                'Correlation',
                'RMSE',
                'R2'
            ]].to_csv(output_dir + 'candidates_reduced.csv')
    
    
    def plot_to_pdf(
        self,
        output_dir = './',
        bin_widths_1d = None,
        bin_edges_2d = None,
        plot_logy = False,
        plot_logx = False,
        sampling_95quantile = False
    ):
        '''
        Plots all candidate functions to pdf files.
        1) Candidate functions -> candidates.pdf.
        2) Candidate functions (sampling of parameters) -> candidates_sampling.pdf.
        3) Goodness-of-fit scores -> candidates_gof.pdf.
        4) Correlation matrices -> candidates_correlation.pdf.
        
        Arguments
        ---------
        output_dir : str
            Output directory.
            
        bin_widths_1d : list | ndarray
            Bin widths for x for plotting 1D histogram data.
            Shape is (num_examples, 1).
        
        bin_edges_2d : list | ndarray
            Bin edges for x for plotting 2D histogram data,
            i.e., [[x0_0, x0_1,...], [x1_0, x1_1,...]],
            where the leftmost bin in x0 has edges x0_0 and x0_1.
            Shape is (num_x0_bins + 1, num_x1_bins + 1).
        
        plot_logy : bool
            Plot functions in log scale for y in candidates.pdf.
            
        plot_logx : bool
            Plot functions in log scale for x in candidates.pdf.
            
        sampling_95quantile : bool
            Whether to include 95% quantile range when plotting
            total uncertainty coverage from ensemble of functions
            generated by sampling parameters in candidates_sampling.pdf.
            If False, plot only the 68% quantile range.
        '''
        
        
        x = self.x
        y = self.y
        y_up = self.y_up
        y_down = self.y_down
        fit_y_unc = self.fit_y_unc
        func_candidates = self.func_candidates
        
        x, y, y_up, y_down, _, dim = dataset_formatting(x = x,
                                                        y = y,
                                                        y_up = y_up,
                                                        y_down = y_down,
                                                        fit_y_unc = fit_y_unc
                                                        )
        
        # Define output directory to store csv and pdf files.
        if output_dir.endswith('/'):
            output_dir = output_dir
            
        else:
            output_dir = output_dir + '/'
            
        os.makedirs(output_dir) if not os.path.exists(output_dir) else None
        
        # Plot results and write to output pdf files.
        if dim == 1:
            plot_all_syst_all_func_1D(func_candidates = func_candidates,
                                      x = x,
                                      bin_widths_1d = bin_widths_1d,
                                      y = y,
                                      y_up = y_up,
                                      y_down = y_down,
                                      pdf_path = output_dir + 'candidates.pdf',
                                      logy = plot_logy,
                                      logx = plot_logx
                                      )
                                      
            plot_total_unc_coverage_all_func_1D(func_candidates = func_candidates,
                                                x = x,
                                                bin_widths_1d = bin_widths_1d,
                                                y = y,
                                                y_up = y_up,
                                                y_down = y_down,
                                                n_samples = 2000,
                                                sampling_95quantile = sampling_95quantile,
                                                pdf_path = output_dir + 'candidates_sampling.pdf',
                                                logy = plot_logy,
                                                logx = plot_logx
                                                )
                                      
        elif dim == 2:
            plot_all_syst_all_func_2D(func_candidates = func_candidates,
                                      x = x,
                                      bin_edges_2d = bin_edges_2d,
                                      y = y,
                                      y_up = y_up,
                                      y_down = y_down,
                                      pdf_path = output_dir + 'candidates.pdf',
                                      logy = plot_logy,
                                      logx = plot_logx
                                      )
        
        plot_all_corr(func_candidates = func_candidates,
                      y_up = y_up,
                      y_down = y_down,
                      pdf_path = output_dir + 'candidates_correlation.pdf'
                      )
        
        plot_all_gof(func_candidates = func_candidates,
                     y_up = y_up,
                     y_down = y_down,
                     pdf_path = output_dir + 'candidates_gof.pdf'
                     )
        
        
    def print_candidate(
        self,
        candidate_number = 99,
    ):
        '''
        Print candidate functions in prompt.
        
        Arguments
        ---------
        candidate_number : int
            Print result for a particular candidate function by setting it to its #,
            or for all candidates by setting it to 99.
        '''
        
        pd.set_option('display.max_colwidth', 30)
        
        # Whether to print relevant info or full info including all intermediate results.
        try:
            func_candidates = self.func_candidates[[
                'Parameterized equation, unscaled',
                'Parameters: (best-fit, +1, -1)',
                'Covariance',
                'Correlation',
                'RMSE',
                'R2',
                'NDF',
                'Chi2',
                'Chi2/NDF',
                'p-value'
            ]]
            
        except:
            # 'NDF', 'Chi2', 'Chi2/NDF' do not exist if y_up/y_down were not considered in fits.
            func_candidates = self.func_candidates[[
                'Parameterized equation, unscaled',
                'Parameters: (best-fit, +1, -1)',
                'Covariance',
                'Correlation',
                'RMSE',
                'R2'
            ]]
        
        # A function to print a particular candidate.
        def print_cand(func_candidate):
        
            print(func_candidate)
            
            # Print candidate function separately with its best-fit parameters and +/-1 sigma substituted.
            # First print the parameterized function before substitution,
            # as well as the parameters and their correlation.
            func_unsub = func_candidate['Parameterized equation, unscaled']
            
            print('\nFunction:\n' + func_unsub)
            
            print('\nParameters (best-fit, +1, -1):\n' + str(func_candidate['Parameters: (best-fit, +1, -1)']))
            
            print('\nCovariance:\n' + str(func_candidate['Covariance']))
            
            print('\nCorrelation:\n' + str(func_candidate['Correlation']))
            
            # Then print the substituted function.
            if len(func_candidate['Parameters: (best-fit, +1, -1)']) > 0:
                # Substitute all parameters with their best-fit values and print the function.
                func_sub = func_unsub
                
                for i in range(len(func_candidate['Parameters: (best-fit, +1, -1)'])):
                    func_sub = func_sub.replace(f'a{i+1}',
                                                str(func_candidate['Parameters: (best-fit, +1, -1)'][f'a{i+1}'][0])
                                                )
                    
                print('\nBest-fit:\n' + func_sub + '\n')
                
                # Substitute with each parameter shifted by +/-1 sigma while keeping others in their best-fit values.
                for i in range(len(func_candidate['Parameters: (best-fit, +1, -1)'])):
                    func_sub = func_unsub
                    
                    # Possible only if the parameter has +/-1 sigma values stored.
                    if func_candidate['Parameters: (best-fit, +1, -1)'][f'a{i+1}'][1] > 0:
                        # Substitute with +/-1 sigma separately for the current parameter.
                        up = func_candidate['Parameters: (best-fit, +1, -1)'][f'a{i+1}'][0]
                        up = up + func_candidate['Parameters: (best-fit, +1, -1)'][f'a{i+1}'][1]
                        
                        down = func_candidate['Parameters: (best-fit, +1, -1)'][f'a{i+1}'][0]
                        down = down + func_candidate['Parameters: (best-fit, +1, -1)'][f'a{i+1}'][2]
                        
                        func_sub_up = func_sub.replace(f'a{i+1}',
                                                       str(round_a_number(up, 6))
                                                       )
                        
                        func_sub_down = func_sub.replace(f'a{i+1}',
                                                         str(round_a_number(down, 6))
                                                         )
                        
                        # Substitute the rest parameters with their best-fit values.
                        for j in range(len(func_candidate['Parameters: (best-fit, +1, -1)'])):
                            # The current parameter i is already substituted with its +/-1 sigma.
                            if j is not i:
                                func_sub_up = func_sub_up.replace(f'a{j+1}',
                                                                  str(func_candidate['Parameters: (best-fit, +1, -1)'][f'a{j+1}'][0])
                                                                  )
                                
                                func_sub_down = func_sub_down.replace(f'a{j+1}',
                                                                      str(func_candidate['Parameters: (best-fit, +1, -1)'][f'a{j+1}'][0])
                                                                      )
                                
                        print('a{} (up):\n'.format(str(i + 1)) + func_sub_up + '\n')
                        
                        print('a{} (down):\n'.format(str(i + 1)) + func_sub_down + '\n')

        
        # Print result for all candidate functions (# = 99) or just for a particular one.
        if candidate_number == 99:
            for i in range(len(func_candidates)):
                print('\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n')
                
                print_cand(func_candidate = func_candidates.iloc[i])
                
                print('<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\n')
        else:
            if candidate_number < len(func_candidates):
                print('\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n')
                
                print_cand(func_candidate = func_candidates.iloc[candidate_number])
                
                print('<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\n')
            else:
                print('Error: candidate_number must be 0-{}'.format(str(len(func_candidates) - 1)))
                
                 



import numpy as np
from numpy import sin, cos, tan, exp, sinh, cosh, tanh, log10, log, where
from numpy import arcsin, arccos, arctan, arcsinh, arccosh, arctanh


def square(x):
    return x*x
    
def cond(x, y):
    return where(x > 0., y, 0.)
    
def piecewise(y, x):
    return where(x > 0., y, 0.)
    
def gauss(x):
    return exp(-x*x)
    
def sigmoid(x):
    return 1./(1.+exp(-x))




import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.image as mpimg
import seaborn as sns
import textwrap
import warnings
warnings.filterwarnings("ignore", category=UserWarning, module="seaborn")
from .utils import *
from .math_defs import *
from .evaluate import *
import scipy
from scipy.interpolate import griddata
from importlib import resources


'''
1D candidate plots
'''

def plot_single_syst_single_func_1D(
    func_candidate,
    candidate_idx,
    x,
    bin_widths_1d,
    y,
    y_up,
    y_down,
    param_shifted,
    logy,
    logx
):
    '''
    Plot a particular candidate function with all parameters at their best-fit values,
    or with one of the parameters in +1 or -1 sigma.
    
    Arguments
    ---------
    func_candidate (pd.dataframe):
        A particular candidate function (one row of the full func_candidates).
        
    candidate_idx (np.int):
        Candidate function # (ranked by function complexity).
    
    x (np.ndarray):
        The independent variable.
        
    y (np.ndarray):
        The dependent variable.
    
    y_up (np.ndarray):
        +1 sigma of y.
        
    y_down (np.ndarray):
        -1 sigma of y.
        
    param_shifted (str):
        Plot the function with a particular parameter shifted (e.g., 'a3').
    
    logy (bool):
        Plot y in log scale.
    
    logx (bool):
        Plot x in log scale.
    '''
    
    fig, axes = plt.subplots(3, 1, sharex=True, figsize=(9,8), gridspec_kw={'height_ratios': [3,1,1]})
    fig.subplots_adjust(hspace = 0.1)
    
    linewidth = 1
    alpha = 0.8
    
    x0 = np.arange(np.min(x), np.max(x), np.abs(np.max(x)-np.min(x))/200)
    
    # Function evaluated with all parameters in their best-fit values (with finer x0).
    central = func_evaluate(func_candidate = func_candidate,
                            x = x0,
                            dim = 1,
                            param_shifted = None,
                            sigma_pm = None
                            )
    
    # Evaluated with one of the parameters shifted up and down separately.
    if param_shifted is not None:
        if func_candidate['Parameters: (best-fit, +1, -1)'][param_shifted][1] > 0:
            up = func_evaluate(func_candidate = func_candidate,
                               x = x0,
                               dim = 1,
                               param_shifted = param_shifted,
                               sigma_pm = '+'
                               )
            
            down = func_evaluate(func_candidate = func_candidate,
                                 x = x0,
                                 dim = 1,
                                 param_shifted = param_shifted,
                                 sigma_pm = '-'
                                 )
    
    # Function evaluated with all parameters in their best-fit values (with the same input x grid).
    # For computing the errors comparing with the input data points directly.
    central_hist = func_evaluate(func_candidate = func_candidate,
                                 x = x,
                                 dim = 1,
                                 param_shifted = None,
                                 sigma_pm = None
                                 )
    
    # Function evaluated with one of the parameters shifted up and down separately (with the same input x grid).
    if param_shifted is not None:
        if func_candidate['Parameters: (best-fit, +1, -1)'][param_shifted][1] > 0:
            up_hist = func_evaluate(func_candidate = func_candidate,
                                    x = x,
                                    dim = 1,
                                    param_shifted = param_shifted,
                                    sigma_pm = '+'
                                    )
            
            down_hist = func_evaluate(func_candidate = func_candidate,
                                      x = x,
                                      dim = 1,
                                      param_shifted = param_shifted,
                                      sigma_pm = '-'
                                      )
    
    
    # Plot the input data in errorbar style if there are uncertainties in y provided,
    # otherwise plot in scatter style.
    if bin_widths_1d is not None:
        bin_widths_1d = np.reshape(np.array(bin_widths_1d), (-1, 1))
        
    if y_up is not None and y_down is not None:
        if bin_widths_1d is not None:
            axes[0].errorbar(x.flatten(),
                             y.flatten(),
                             yerr = [y_down.flatten(), y_up.flatten()],
                             xerr = bin_widths_1d.flatten() / 2,
                             fmt = '.',
                             c = 'black',
                             ecolor = 'grey',
                             capsize = 0,
                             label = 'Data'
                             )
            
        else:
            axes[0].errorbar(x.flatten(),
                             y.flatten(),
                             yerr = [y_down.flatten(), y_up.flatten()],
                             fmt = '.',
                             c = 'black',
                             ecolor = 'grey',
                             capsize = 0,
                             label = 'Data'
                             )
    else:
        axes[0].scatter(x,
                        y,
                        marker = 'o',
                        c = 'black',
                        alpha = 1,
                        label = 'Data'
                        )
                        
    # Plot the candidate function with all parameters in their best-fit values.
    axes[0].plot(x0,
                 central,
                 color = 'r',
                 label = 'Best-fit',
                 linewidth = linewidth,
                 alpha = alpha
                 )
    
    # Plot the candidate function with one of the parameters shifted up and down separately.
    if param_shifted is not None:
        if func_candidate['Parameters: (best-fit, +1, -1)'][param_shifted][1] > 0:
            axes[0].plot(x0,
                         up,
                         color = 'g',
                         label = '{0} Up'.format(param_shifted),
                         linewidth = linewidth,
                         alpha = alpha,
                         linestyle = 'dashed'
                         )
            
            axes[0].plot(x0,
                         down,
                         color = 'b',
                         label = '{0} Down'.format(param_shifted),
                         linewidth = linewidth,
                         alpha = alpha,
                         linestyle = 'dashed'
                         )
            
    axes[0].legend()
    
    # Define a string containing all the parameters in the function with each one in the form of
    # best-fit^{+1 sigma}_{-1 sigma}, with relative % error next to the +/-1 sigma values.
    # Highlight the one that is being shifted.
    parameters_string = ""
    parameters = func_candidate['Parameters: (best-fit, +1, -1)']
    
    for i in range(len(parameters)):
        # If the parameter is held fixed in the fit, it does not have +/-1 sigma values.
        if parameters[f'a{i+1}'][1] == 0:
            if f'a{i+1}' == param_shifted:
                parameters_string += r"$\bf{{{0} = {1}}}$".format(
                    f'a{i+1}',
                    round_a_number(parameters[f'a{i+1}'][0], 6),
                    )
                    
            else:
                parameters_string += r"${0} = {1}$".format(
                    f'a{i+1}',
                    round_a_number(parameters[f'a{i+1}'][0], 6),
                    )
        else:
            if f'a{i+1}' == param_shifted:
                parameters_string += r"$\bf{{{0} = {1}^{{+ {2} ({4}\%) }}_{{- {3} ({5}\%)}}}}$".format(
                    f'a{i+1}',
                    round_a_number(parameters[f'a{i+1}'][0], 6),
                    round_a_number(parameters[f'a{i+1}'][1], 4),
                    round_a_number(np.abs(parameters[f'a{i+1}'][2]), 4),
                    round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][1]) / float(parameters[f'a{i+1}'][0])), 3),
                    round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][2]) / float(parameters[f'a{i+1}'][0])), 3)
                    )
                    
            else:
                parameters_string += r"$\text{{{0}}} = {1}^{{+ {2} ({4}\%) }}_{{- {3} ({5}\%)}}$".format(
                f'a{i+1}',
                round_a_number(parameters[f'a{i+1}'][0], 6),
                round_a_number(parameters[f'a{i+1}'][1], 4),
                round_a_number(np.abs(parameters[f'a{i+1}'][2]), 4),
                round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][1]) / float(parameters[f'a{i+1}'][0])), 3),
                round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][2]) / float(parameters[f'a{i+1}'][0])), 3)
                )
        
        if i < len(parameters) - 1:
            parameters_string += ",  "
            
        if np.mod(i + 1, 2) == 0:
            parameters_string += "\n"
            
    # Print parameter string defined above as title.
    title = textwrap.fill(func_candidate['Parameterized equation, unscaled'], width = 95) + "\n\n" + parameters_string
    axes[0].set_title(title, loc = 'left', size = 9.5)
    
    # Show the candidate # being plotted and some of the gof metrics.
    if y_up is not None and y_down is not None:
        axes[0].set_title(r"$\bfit{{Candidate\,\#{}}}$".format(candidate_idx) + "\n$\\chi^2$/NDF = {0}/{1}, p-value = {3}, RMSE = {2}".format(func_candidate['Chi2'], func_candidate['NDF'], func_candidate['RMSE'], round_a_number(func_candidate['p-value'], 4)), loc = 'right', size = 9.5)
        
    else:
        axes[0].set_title(r"$\bfit{{Candidate\,\#{}}}$".format(candidate_idx) + "\nRMSE = {0}, R2 = {1}".format(func_candidate['RMSE'], func_candidate['R2']), loc = 'right', size = 9.5)

    formatter = ticker.FormatStrFormatter('%.3g')
    
    # Plot a panel showing the difference between the prediction and the input data point, essentially (y_pred - y_label) in sigmas.
    axes[1].plot(x0,
                 np.zeros(central.shape),
                 c = 'black',
                 linestyle = 'dashed',
                 linewidth = linewidth,
                 alpha = 0.5
                 )
    
    if y_up is not None and y_down is not None:
        residual_central = y - central_hist
        
        # Plot the residual error in terms of the input y uncertainties.
        y_unc_central =  np.where(residual_central < 0,
                                  np.where(y_up != 0, y_up, y_down),
                                  np.where(y_down != 0, y_down, y_up)
                                  )
                                  
        if bin_widths_1d is not None:
            axes[1].bar(x.flatten(),
                        (residual_central/y_unc_central).flatten(),
                        width=bin_widths_1d.flatten(),
                        edgecolor='none',
                        color='red',
                        alpha=alpha
                        )
        else:
            axes[1].scatter(x,
                            residual_central/y_unc_central,
                            marker='.',
                            c='r',
                            alpha=alpha
                            )
        
        axes[1].set_ylim(-1.3*max(np.abs(residual_central/y_unc_central)),
                         1.3*max(np.abs(residual_central/y_unc_central))
                         )
        
        # Draw reference lines showing +/-1 sigma.
        #axes[1].plot(x0, np.ones(central.shape).flatten(), c='grey', marker='none', alpha=0.7)
        #axes[1].plot(x0, -np.ones(central.shape).flatten(), c='grey', marker='none', alpha=0.7)
        #axes[1].fill_between(x0.flatten(), np.ones(central.shape).flatten(), -np.ones(central.shape).flatten(), color='grey', alpha=0.15, rasterized=True)
    else:
        if bin_widths_1d is not None:
            axes[1].bar(x.flatten(),
                        (y - central_hist).flatten(),
                        width=bin_widths_1d.flatten(),
                        edgecolor='none',
                        color='red',
                        alpha=alpha
                        )
        
        else:
            axes[1].scatter(x,
                            y - central_hist,
                            marker='.',
                            c='r',
                            alpha=alpha)
                            
            
        axes[1].set_ylim(-1.3*max(np.abs(y - central_hist)),
                         1.3*max(np.abs(y - central_hist))
                         )
    
    # Same residual plot as above but for the function evaluated with one of the parameters shifted.
    if param_shifted is not None:
        if func_candidate['Parameters: (best-fit, +1, -1)'][param_shifted][1] > 0:
            if y_up is not None and y_down is not None:
                residual_up = y - up_hist
                
                y_unc_up =  np.where(residual_up < 0,
                                     np.where(y_up != 0, y_up, y_down),
                                     np.where(y_down != 0, y_down, y_up))
                                     
                residual_down = y - down_hist
                
                y_unc_down =  np.where(residual_down < 0,
                                       np.where(y_up != 0, y_up, y_down),
                                       np.where(y_down != 0, y_down, y_up))
                                       
                #axes[1].scatter(x, residual_up/y_unc_up, marker='.', c='g', alpha=alpha)
                #axes[1].scatter(x, residual_down/y_unc_down, marker='.', c='b', alpha=alpha)
            #else:
                #axes[1].scatter(x, y - up_hist, marker='.', c='g', alpha=alpha)
                #axes[1].scatter(x, y - down_hist, marker='.', c='b', alpha=alpha)
                
    if y_up is not None and y_down is not None:
        #axes[1].plot(x, y_up, marker='none', c='grey', alpha=0.3)
        #axes[1].plot(x, -y_down, marker='none', c='grey', alpha=0.3)
        #axes[1].fill_between(x.flatten(), y_up.flatten(), -y_down.flatten(), color='grey', alpha=0.15)
    #axes[1].legend(loc='upper right', bbox_to_anchor=(1.26, 1))
        axes[1].set_ylabel('$\\frac{\\text{Data}-\\text{Fit}}{\\text{Data unc.}}$', fontsize=15)
        
    else:
        axes[1].set_ylabel('Data$-$Fit')
        
    axes[1].yaxis.set_label_position("right")
    axes[1].yaxis.set_major_formatter(formatter)
    
    # Ratio plot: (function evaluated with one of the parameters shifted) / (function evaluated with all parameters in best-fit).
    # This is to show the uncertainty variations comparing with the central best-fit.
    axes[2].plot(x0, np.ones(central.shape), color='black', linestyle='dashed', linewidth=linewidth, alpha=0.5)
    
    if param_shifted is not None:
        if func_candidate['Parameters: (best-fit, +1, -1)'][param_shifted][1] > 0:
            axes[2].plot(x0,
                         up/central,
                         color = 'g',
                         linewidth = linewidth,
                         alpha = alpha
                         )
                         
            axes[2].plot(x0,
                         down/central,
                         color = 'b',
                         linewidth = linewidth,
                         alpha = alpha
                         )
                         
    #axes[2].legend(loc='upper right', bbox_to_anchor=(1.25, 1))
    axes[2].set_ylabel('$\\frac{\\text{Up or Down}}{\\text{Best-fit}}$', fontsize=15)
    
    axes[2].yaxis.set_label_position("right")
    axes[2].yaxis.set_major_formatter(formatter)
    
    #axes[2].set_ylim(0.5,1.5)
    
    if logy:
        axes[0].set_yscale('log')
        #axes[1].set_yscale('symlog')
        
    if logx:
        axes[0].set_xscale('log')
    
    # Add logo to every plot.
    logo_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'img/logo.png'
                             )
    logo_img = mpimg.imread(logo_path)
    
    bbox = axes[0].get_position()
    
    logo_width = 0.13
    logo_height = logo_width * (logo_img.shape[0] / logo_img.shape[1])

    ax_inset = fig.add_axes([bbox.x0 +0.73,
                            bbox.y1 + 0.085,
                            logo_width,
                            logo_height]
                            )

    ax_inset.imshow(logo_img)
    ax_inset.axis('off')
    
    plt.tight_layout()
                

def plot_all_syst_all_func_1D(
    func_candidates,
    x,
    bin_widths_1d,
    y,
    y_up,
    y_down,
    pdf_path,
    logy,
    logx
):
    '''
    Plot all candidate functions, each with all possible parameter variations.
    
    Arguments
    ---------
    func_candidates (pd.dataframe):
        Full dataframe containing all candidate functions after fits.
    
    x (np.ndarray):
        The independent variable.
        
    y (np.ndarray):
        The dependent variable.
    
    y_up (np.ndarray):
        +1 sigma of y.
        
    y_down (np.ndarray):
        -1 sigma of y.
        
    pdf_path (str):
        Save the output files to this directory.
    
    logy (bool):
        Plot y in log scale.
    
    logx (bool):
        Plot x in log scale.
    
    
    Returns
    -------
    Plot all candidate functions with all parameter variations and write to an output file.
    '''
    
    with PdfPages(pdf_path) as pdf:
        for i in range(len(func_candidates)):
            # Print candidate # page.
            fig, axes = plt.subplots(figsize = (9, 7))
            axes.axis('off')
            plt.text(0.5,
                     0.5,
                     'Candidate function #{}'.format(len(func_candidates) - 1 - i),
                     fontsize = 20,
                     ha = 'center',
                     va = 'center'
                     )
                     
            plt.tight_layout()
            
            plt.savefig(pdf, format = 'pdf')
            
            plt.close()
            
            if i < len(func_candidates) - 1:
                print('Plotting candidate functions {0}/{1} >>> {2}'.format(i+1, len(func_candidates), pdf_path), end='\r')
                
            else:
                print('Plotting candidate functions {0}/{1} >>> {2}'.format(i+1, len(func_candidates), pdf_path))
                
            func_candidate = func_candidates.iloc[len(func_candidates) - 1 - i]
            
            if len(func_candidate['Parameters: (best-fit, +1, -1)']) > 0:
                # Check if the function has any parameter that has variations.
                has_uncert = False
                
                for j in range(len(func_candidate['Parameters: (best-fit, +1, -1)'])):
                    if func_candidate['Parameters: (best-fit, +1, -1)'][f'a{j+1}'][1] > 0:
                        has_uncert = True

                # Plot central, up, and down in the same plot, otherwise plot only the central.
                if has_uncert:
                    for j in range(len(func_candidate['Parameters: (best-fit, +1, -1)'])):
                        if func_candidate['Parameters: (best-fit, +1, -1)'][f'a{j+1}'][1] > 0:
                            plot_single_syst_single_func_1D(func_candidate = func_candidate,
                                                            candidate_idx = len(func_candidates) - 1 - i,
                                                            x = x,
                                                            bin_widths_1d = bin_widths_1d,
                                                            y = y,
                                                            y_up = y_up,
                                                            y_down = y_down,
                                                            param_shifted = f'a{j+1}',
                                                            logy = logy,
                                                            logx = logx
                                                            )
                            
                            plt.savefig(pdf, format='pdf')
                            plt.close()
                            
                else:
                    plot_single_syst_single_func_1D(func_candidate = func_candidate,
                                                    candidate_idx = len(func_candidates) - 1 - i,
                                                    x = x,
                                                    bin_widths_1d = bin_widths_1d,
                                                    y = y,
                                                    y_up = y_up,
                                                    y_down = y_down,
                                                    param_shifted = None,
                                                    logy = logy,
                                                    logx = logx
                                                    )
                    
                    plt.savefig(pdf, format='pdf')
                    plt.close()
                    
            else:
                plot_single_syst_single_func_1D(func_candidate = func_candidate,
                                                candidate_idx = len(func_candidates) - 1 - i,
                                                x = x,
                                                bin_widths_1d = bin_widths_1d,
                                                y = y,
                                                y_up = y_up,
                                                y_down = y_down,
                                                param_shifted = None,
                                                logy = logy,
                                                logx = logx
                                                )
                
                plt.savefig(pdf, format='pdf')
                plt.close()
                
                
'''
Monte Carlo sampling
'''

def plot_total_unc_coverage_single_func_1D(
    func_candidate,
    candidate_idx,
    x,
    bin_widths_1d,
    y,
    y_up,
    y_down,
    n_samples,
    sampling_95quantile,
    logy,
    logx
):
    '''
    Plot a particular candidate function with total uncertainty coverage.
    The coverage is computed from an ensemble of functions where the parameters
    are sampled from the multivariate normal distribution according to the
    best-fit parameter values and the covariance matrix.
    
    Arguments
    ---------
    func_candidate (pd.dataframe):
        A particular candidate function (one row of the full func_candidates).
        
    candidate_idx (np.int):
        Candidate function # (ranked by function complexity).
    
    x (np.ndarray):
        The independent variable.
        
    y (np.ndarray):
        The dependent variable.
    
    y_up (np.ndarray):
        +1 sigma of y.
        
    y_down (np.ndarray):
        -1 sigma of y.
        
    n_samples (int):
        Number of samples to be drawn from the multivariate normal distribution.
    
    logy (bool):
        Plot y in log scale.
    
    logx (bool):
        Plot x in log scale.
    '''
    fig, axes = plt.subplots(3, 1, sharex=True, figsize=(9,8), gridspec_kw={'height_ratios': [3,1,1]})
    fig.subplots_adjust(hspace = 0.1)
    
    linewidth = 1
    alpha = 0.8
    
    x0 = np.arange(np.min(x), np.max(x), np.abs(np.max(x)-np.min(x))/200)
    
    if n_samples is not None:
        func_bands, func_bands_finer = func_sampling_1d(func_str = func_candidate['Parameterized equation, unscaled'],
                                                                 param = func_candidate['Parameters: (best-fit, +1, -1)'],
                                                                 cov = func_candidate['Covariance'],
                                                                 x = x,
                                                                 x_finer = x0,
                                                                 n_samples = n_samples
                                                                 )
                                                                 
        (mean_func, lower_2sigma, lower_1sigma, upper_1sigma, upper_2sigma) = func_bands
        (mean_func_finer, lower_2sigma_finer, lower_1sigma_finer, upper_1sigma_finer, upper_2sigma_finer) = func_bands_finer
        
        axes[0].plot(x0, mean_func_finer, label = 'Sample mean', color = 'red')
        
        axes[0].fill_between(x0,
                             lower_1sigma_finer,
                             upper_1sigma_finer,
                             color = 'limegreen',
                             alpha = 0.5,
                             label = '68% quantile range'
                             )
        
        if sampling_95quantile:
            axes[0].fill_between(x0,
                                 lower_2sigma_finer,
                                 lower_1sigma_finer,
                                 color = 'gold',
                                 alpha = 0.5
                                 )
                             
            axes[0].fill_between(x0,
                                 upper_1sigma_finer,
                                 upper_2sigma_finer,
                                 color = 'gold',
                                 alpha = 0.5,
                                 label = '95% quantile range'
                                 )
                             
    else:
        # Function evaluated with all parameters in their best-fit values (with finer x0).
        central = func_evaluate(func_candidate = func_candidate,
                                x = x0,
                                dim = 1,
                                param_shifted = None,
                                sigma_pm = None
                                )
                                
        # Function evaluated with all parameters in their best-fit values (with the same input x grid).
        # For computing the errors comparing with the input data points directly.
        central_hist = func_evaluate(func_candidate = func_candidate,
                                    x = x,
                                    dim = 1,
                                    param_shifted = None,
                                    sigma_pm = None
                                    )
                                    
        axes[0].plot(x0,
                    central,
                    color = 'r',
                    label = 'Best-fit',
                    linewidth = linewidth,
                    alpha = alpha
                    )
    
    # Plot the input data in errorbar style if there are uncertainties in y provided,
    # otherwise plot in scatter style.
    if bin_widths_1d is not None:
        bin_widths_1d = np.reshape(np.array(bin_widths_1d), (-1, 1))
        
    if y_up is not None and y_down is not None:
        if bin_widths_1d is not None:
            axes[0].errorbar(x.flatten(),
                             y.flatten(),
                             yerr = [y_down.flatten(), y_up.flatten()],
                             xerr = bin_widths_1d.flatten() / 2,
                             fmt = '.',
                             c = 'black',
                             ecolor = 'grey',
                             capsize = 0,
                             label = 'Data'
                             )
            
        else:
            axes[0].errorbar(x.flatten(),
                             y.flatten(),
                             yerr = [y_down.flatten(), y_up.flatten()],
                             fmt = '.',
                             c = 'black',
                             ecolor = 'grey',
                             capsize = 0,
                             label = 'Data'
                             )
    else:
        axes[0].scatter(x,
                        y,
                        marker = 'o',
                        c = 'black',
                        alpha = 1,
                        label = 'Data'
                        )
                        
            
    axes[0].legend()
    
    # Define a string containing all the parameters in the function with each one in the form of
    # best-fit^{+1 sigma}_{-1 sigma}, with relative % error next to the +/-1 sigma values.
    # Highlight the one that is being shifted.
    parameters_string = ""
    parameters = func_candidate['Parameters: (best-fit, +1, -1)']
    
    for i in range(len(parameters)):
        # If the parameter is held fixed in the fit, it does not have +/-1 sigma values.
        if parameters[f'a{i+1}'][1] == 0:
            parameters_string += r"${0} = {1}$".format(
                f'a{i+1}',
                round_a_number(parameters[f'a{i+1}'][0], 6),
                )
                
        else:
            parameters_string += r"$\text{{{0}}} = {1}^{{+ {2} ({4}\%) }}_{{- {3} ({5}\%)}}$".format(
            f'a{i+1}',
            round_a_number(parameters[f'a{i+1}'][0], 6),
            round_a_number(parameters[f'a{i+1}'][1], 4),
            round_a_number(np.abs(parameters[f'a{i+1}'][2]), 4),
            round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][1]) / float(parameters[f'a{i+1}'][0])), 3),
            round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][2]) / float(parameters[f'a{i+1}'][0])), 3)
            )
        
        if i < len(parameters) - 1:
            parameters_string += ",  "
            
        if np.mod(i + 1, 2) == 0:
            parameters_string += "\n"
            
    # Print parameter string defined above as title.
    title = textwrap.fill(func_candidate['Parameterized equation, unscaled'], width = 95) + "\n\n" + parameters_string
    axes[0].set_title(title, loc = 'left', size = 9.5)
    
    # Show the candidate # being plotted and some of the gof metrics.
    if n_samples is not None:
        axes[0].set_title(r"$\bfit{{Candidate\,\#{}}}$".format(candidate_idx) + "\nEnsemble of functions generated by sampling parameters", loc = 'right', size = 9.5)
    else:
        axes[0].set_title(r"$\bfit{{Candidate\,\#{}}}$".format(candidate_idx) + "\n", loc = 'right', size = 9.5)

    formatter = ticker.FormatStrFormatter('%.3g')
    
    # Plot a panel showing the difference between the prediction and the input data point, essentially (y_pred - y_label) in sigmas.
    axes[1].plot(x0,
                 np.zeros(x0.shape),
                 c = 'black',
                 linestyle = 'dashed',
                 linewidth = linewidth,
                 alpha = 0.5
                 )
    
    if y_up is not None and y_down is not None:
        if n_samples is not None:
            residual_mean = y - mean_func
            residual_upper_1sigma = y - upper_1sigma
            residual_lower_1sigma = y - lower_1sigma
            
            if sampling_95quantile:
                residual_upper_2sigma = y - upper_2sigma
                residual_lower_2sigma = y - lower_2sigma
            
        else:
            residual_mean = y - central_hist
        
        # Plot the residual error in terms of the input y uncertainties.
        y_unc_mean =  np.where(residual_mean < 0,
                               np.where(y_up != 0, y_up, y_down),
                               np.where(y_down != 0, y_down, y_up)
                               )
                               
        if n_samples is not None:
            y_unc_upper_1sigma = np.where(residual_upper_1sigma < 0,
                                          np.where(y_up != 0, y_up, y_down),
                                          np.where(y_down != 0, y_down, y_up)
                                          )
                                          
            y_unc_lower_1sigma = np.where(residual_lower_1sigma < 0,
                                          np.where(y_up != 0, y_up, y_down),
                                          np.where(y_down != 0, y_down, y_up)
                                          )
                                          
            if sampling_95quantile:
                y_unc_upper_2sigma = np.where(residual_upper_2sigma < 0,
                                              np.where(y_up != 0, y_up, y_down),
                                              np.where(y_down != 0, y_down, y_up)
                                              )
                                              
                y_unc_lower_2sigma = np.where(residual_lower_2sigma < 0,
                                              np.where(y_up != 0, y_up, y_down),
                                              np.where(y_down != 0, y_down, y_up)
                                              )
                                  
        if bin_widths_1d is not None:
            if n_samples is not None:
                for i in range(x.size):
                    axes[1].hlines((residual_mean/y_unc_mean).flatten()[i],
                                   xmin=x[i] - bin_widths_1d[i] / 2,
                                   xmax=x[i] + bin_widths_1d[i] / 2,
                                   color='red',
                                   linewidth=1.5
                                   )
                
                axes[1].bar(x.flatten(),
                            (residual_lower_1sigma/y_unc_lower_1sigma - residual_mean/y_unc_mean).flatten(),
                            width = bin_widths_1d.flatten(),
                            bottom = (residual_mean/y_unc_mean).flatten(),
                            color = 'limegreen',
                            edgecolor = 'none',
                            alpha = 0.5,
                            )
                            
                axes[1].bar(x.flatten(),
                            (residual_upper_1sigma/y_unc_upper_1sigma - residual_mean/y_unc_mean).flatten(),
                            width = bin_widths_1d.flatten(),
                            bottom = (residual_mean/y_unc_mean).flatten(),
                            color = 'limegreen',
                            edgecolor = 'none',
                            alpha = 0.5,
                            )
                            
                if sampling_95quantile:
                    axes[1].bar(x.flatten(),
                                (- residual_lower_1sigma/y_unc_lower_1sigma + residual_lower_2sigma/y_unc_lower_2sigma).flatten(),
                                width = bin_widths_1d.flatten(),
                                bottom = (residual_lower_1sigma/y_unc_lower_1sigma).flatten(),
                                color = 'gold',
                                edgecolor = 'none',
                                alpha = 0.5,
                                )
                                
                    axes[1].bar(x.flatten(),
                                (residual_upper_2sigma/y_unc_upper_2sigma - residual_upper_1sigma/y_unc_upper_1sigma).flatten(),
                                width = bin_widths_1d.flatten(),
                                bottom = (residual_upper_1sigma/y_unc_upper_1sigma).flatten(),
                                color = 'gold',
                                edgecolor = 'none',
                                alpha = 0.5,
                                )
            
            else:
                axes[1].bar(x.flatten(),
                            (residual_mean/y_unc_mean).flatten(),
                            width=bin_widths_1d.flatten(),
                            edgecolor='none',
                            color='red',
                            alpha=alpha
                            )
            
                        
        else:
            axes[1].scatter(x,
                            residual_mean/y_unc_mean,
                            marker='.',
                            c='r',
                            alpha=alpha
                            )
        
        if n_samples is not None:
            if sampling_95quantile:
                axes[1].set_ylim(-1.1*np.max(np.maximum(np.abs(residual_lower_2sigma/y_unc_lower_2sigma),
                                                         np.abs(residual_upper_2sigma/y_unc_upper_2sigma))
                                             ),
                                 1.1*np.max(np.maximum(np.abs(residual_lower_2sigma/y_unc_lower_2sigma),
                                                        np.abs(residual_upper_2sigma/y_unc_upper_2sigma))
                                            )
                                )
                                
            else:
                axes[1].set_ylim(-1.1*np.max(np.maximum(np.abs(residual_lower_1sigma/y_unc_lower_1sigma),
                                                         np.abs(residual_upper_1sigma/y_unc_upper_1sigma))
                                             ),
                                 1.1*np.max(np.maximum(np.abs(residual_lower_1sigma/y_unc_lower_1sigma),
                                                        np.abs(residual_upper_1sigma/y_unc_upper_1sigma))
                                            )
                                )
                            
        else:
            axes[1].set_ylim(-1.3*max(np.abs(residual_mean/y_unc_mean)),
                             1.3*max(np.abs(residual_mean/y_unc_mean))
                             )
        
        
    else:
        if n_samples is not None:
            residual_mean = y - mean_func
            
        else:
            residual_mean = y - central_hist
            
        if bin_widths_1d is not None:
            axes[1].bar(x.flatten(),
                        residual_mean.flatten(),
                        width=bin_widths_1d.flatten(),
                        edgecolor='none',
                        color='red',
                        alpha=alpha
                        )
        
        else:
            axes[1].scatter(x,
                            residual_mean,
                            marker='.',
                            c='r',
                            alpha=alpha)
                            
            
        axes[1].set_ylim(-1.3*max(np.abs(residual_mean)),
                         1.3*max(np.abs(residual_mean))
                         )
                
    if y_up is not None and y_down is not None:
        axes[1].set_ylabel('$\\frac{\\text{Data}-\\text{Mean}}{\\text{Data unc.}}$', fontsize=15)
        
    else:
        axes[1].set_ylabel('Data$-$Fit')
        
    axes[1].yaxis.set_label_position("right")
    axes[1].yaxis.set_major_formatter(formatter)
    
    # Ratio plot: (function evaluated with one of the parameters shifted) / (function evaluated with all parameters in best-fit).
    # This is to show the uncertainty variations comparing with the central best-fit.
    axes[2].plot(x0, np.ones(x0.shape), color='black', linestyle='dashed', linewidth=linewidth, alpha=0.5)
    
    if n_samples is not None:
        axes[2].plot(x0,
                     upper_1sigma_finer/mean_func_finer,
                     color = 'limegreen',
                     linewidth = linewidth,
                     alpha = 1
                     )
                     
        axes[2].plot(x0,
                     lower_1sigma_finer/mean_func_finer,
                     color = 'limegreen',
                     linewidth = linewidth,
                     alpha = 1
                     )
                     
        if sampling_95quantile:
            axes[2].plot(x0,
                         upper_2sigma_finer/mean_func_finer,
                         color = 'gold',
                         linewidth = linewidth,
                         alpha = 1
                         )
                         
            axes[2].plot(x0,
                         lower_2sigma_finer/mean_func_finer,
                         color = 'gold',
                         linewidth = linewidth,
                         alpha = 1
                         )
                         
    #axes[2].legend(loc='upper right', bbox_to_anchor=(1.25, 1))
    axes[2].set_ylabel('$\\frac{\\text{Quantile range}}{\\text{Mean}}$', fontsize=15)
    
    axes[2].yaxis.set_label_position("right")
    axes[2].yaxis.set_major_formatter(formatter)
    
    #axes[2].set_ylim(0.5,1.5)
    
    if logy:
        axes[0].set_yscale('log')
        #axes[1].set_yscale('symlog')
        
    if logx:
        axes[0].set_xscale('log')
    
    # Add logo to every plot.
    logo_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'img/logo.png'
                             )
    logo_img = mpimg.imread(logo_path)
    
    bbox = axes[0].get_position()
    
    logo_width = 0.13
    logo_height = logo_width * (logo_img.shape[0] / logo_img.shape[1])

    ax_inset = fig.add_axes([bbox.x0 +0.73,
                            bbox.y1 + 0.085,
                            logo_width,
                            logo_height]
                            )

    ax_inset.imshow(logo_img)
    ax_inset.axis('off')
    
    plt.tight_layout()
                

def plot_total_unc_coverage_all_func_1D(
    func_candidates,
    x,
    bin_widths_1d,
    y,
    y_up,
    y_down,
    n_samples,
    sampling_95quantile,
    pdf_path,
    logy,
    logx
):
    '''
    Plot all candidate functions with total uncertainty coverage.
    The coverage is computed from an ensemble of functions where the parameters
    are sampled from the multivariate normal distribution according to the
    best-fit parameter values and the covariance matrix.
    
    Arguments
    ---------
    func_candidates (pd.dataframe):
        Full dataframe containing all candidate functions after fits.
    
    x (np.ndarray):
        The independent variable.
        
    y (np.ndarray):
        The dependent variable.
    
    y_up (np.ndarray):
        +1 sigma of y.
        
    y_down (np.ndarray):
        -1 sigma of y.
        
    n_samples (int):
        Number of samples to be drawn from the multivariate normal distribution.
        
    pdf_path (str):
        Save the output files to this directory.
    
    logy (bool):
        Plot y in log scale.
    
    logx (bool):
        Plot x in log scale.
    
    
    Returns
    -------
    Plot all candidate functions with total unc coverage and write to an output file.
    '''
    
    with PdfPages(pdf_path) as pdf:
        for i in range(len(func_candidates)):
            # Print candidate # page.
            fig, axes = plt.subplots(figsize = (9, 7))
            axes.axis('off')
            plt.text(0.5,
                     0.5,
                     'Candidate function #{}'.format(len(func_candidates) - 1 - i),
                     fontsize = 20,
                     ha = 'center',
                     va = 'center'
                     )
                     
            plt.tight_layout()
            
            plt.savefig(pdf, format = 'pdf')
            
            plt.close()
            
            if i < len(func_candidates) - 1:
                print('Plotting candidate functions (sampling parameters) {0}/{1} >>> {2}'.format(i+1, len(func_candidates), pdf_path), end='\r')
                
            else:
                print('Plotting candidate functions (sampling parameters) {0}/{1} >>> {2}'.format(i+1, len(func_candidates), pdf_path))
                
            func_candidate = func_candidates.iloc[len(func_candidates) - 1 - i]
            
            if len(func_candidate['Parameters: (best-fit, +1, -1)']) > 0:
                # Check if the function has any parameter that has variations.
                has_uncert = False
                
                for j in range(len(func_candidate['Parameters: (best-fit, +1, -1)'])):
                    if func_candidate['Parameters: (best-fit, +1, -1)'][f'a{j+1}'][1] > 0:
                        has_uncert = True

                if has_uncert:
                    try:
                        plot_total_unc_coverage_single_func_1D(func_candidate = func_candidate,
                                                               candidate_idx = len(func_candidates) - 1 - i,
                                                               x = x,
                                                               bin_widths_1d = bin_widths_1d,
                                                               y = y,
                                                               y_up = y_up,
                                                               y_down = y_down,
                                                               n_samples = n_samples,
                                                               sampling_95quantile = sampling_95quantile,
                                                               logy = logy,
                                                               logx = logx
                                                               )
                                                               
                    except:
                        plot_total_unc_coverage_single_func_1D(func_candidate = func_candidate,
                                                               candidate_idx = len(func_candidates) - 1 - i,
                                                               x = x,
                                                               bin_widths_1d = bin_widths_1d,
                                                               y = y,
                                                               y_up = y_up,
                                                               y_down = y_down,
                                                               n_samples = None,
                                                               sampling_95quantile = False,
                                                               logy = logy,
                                                               logx = logx
                                                               )
                    
                    plt.savefig(pdf, format='pdf')
                    plt.close()
                        
                else:
                    plot_total_unc_coverage_single_func_1D(func_candidate = func_candidate,
                                                           candidate_idx = len(func_candidates) - 1 - i,
                                                           x = x,
                                                           bin_widths_1d = bin_widths_1d,
                                                           y = y,
                                                           y_up = y_up,
                                                           y_down = y_down,
                                                           n_samples = None,
                                                           sampling_95quantile = False,
                                                           logy = logy,
                                                           logx = logx
                                                           )
                    
                    plt.savefig(pdf, format='pdf')
                    plt.close()
                
            else:
                plot_total_unc_coverage_single_func_1D(func_candidate = func_candidate,
                                                       candidate_idx = len(func_candidates) - 1 - i,
                                                       x = x,
                                                       bin_widths_1d = bin_widths_1d,
                                                       y = y,
                                                       y_up = y_up,
                                                       y_down = y_down,
                                                       n_samples = None,
                                                       sampling_95quantile = False,
                                                       logy = logy,
                                                       logx = logx
                                                       )
                
                plt.savefig(pdf, format='pdf')
                plt.close()


'''
2D candidate plots
'''

def plot_single_syst_single_func_2D(
    func_candidate,
    candidate_idx,
    x,
    bin_edges_2d,
    y,
    y_up,
    y_down,
    param_shifted,
    logy,
    logx
):
    '''
    Plot a particular candidate function with all parameters at their best-fit values,
    or with one of the parameters in +1 or -1 sigma.
    
    Arguments
    ---------
    func_candidate (pd.dataframe):
        A particular candidate function (one row of the full func_candidates).
        
    candidate_idx (np.int):
        Candidate function # (ranked by function complexity).
    
    x (np.ndarray):
        The independent variable.
        
    y (np.ndarray):
        The dependent variable.
    
    y_up (np.ndarray):
        +1 sigma of y.
        
    y_down (np.ndarray):
        -1 sigma of y.
        
    param_shifted (str):
        Plot the function with a particular parameter shifted (e.g., 'a3').
    
    logy (bool):
        Plot y in log scale.
    
    logx (bool):
        Plot x in log scale.
    
    
    Returns
    -------
    Plot the candidate function.
    '''
    
    fig, axes = plt.subplots(2, 2, sharex = True, sharey = True, figsize = (9, 7),
                             gridspec_kw = {'width_ratios': [1, 1], 'height_ratios': [1, 1]}
                             )
    #fig.subplots_adjust(hspace = 0.1)
    alpha = 1
    cbar_fontsize = 14
    label_fontsize = 12
    
    x0_bins = np.reshape(np.array(bin_edges_2d[0]), (-1))
    x1_bins = np.reshape(np.array(bin_edges_2d[1]), (-1))
    
    # Plot input
    fig_data = axes[0,0].hist2d(x[:,0],
                                x[:,1],
                                bins = [x0_bins, x1_bins],
                                weights = np.squeeze(y),
                                cmap = 'Greens',
                                edgecolor = 'none',
                                rasterized = True
                                )
                                
    cbar_data = plt.colorbar(fig_data[3], ax=axes[0,0], pad=0, label='Data')
    
    cbar_data.ax.yaxis.label.set_size(cbar_fontsize)
    
    axes[0,0].set_xlabel('x0', fontsize = label_fontsize)
    axes[0,0].set_ylabel('x1', fontsize = label_fontsize)

    
    # Evaluate the candidate function in finer bins
    x0_nbins = 200
    x1_nbins = 200
    
    x0_range = np.linspace(min(x0_bins), max(x0_bins), x0_nbins)
    x1_range = np.linspace(min(x1_bins), max(x1_bins), x1_nbins)
    
    x0_smooth, x1_smooth = np.meshgrid(x0_range, x1_range)
    
    x_smooth = np.column_stack((x0_smooth.flatten(), x1_smooth.flatten()))
    
    central_smooth = func_evaluate(func_candidate = func_candidate,
                                   x = x_smooth,
                                   dim = 2,
                                   param_shifted = None,
                                   sigma_pm = None
                                   )
    
    # Plot the candidate function (smooth)
    fig_fitted_smooth = axes[0,1].hist2d(x_smooth[:,0],
                                         x_smooth[:,1],
                                         bins = (x0_nbins, x1_nbins),
                                         weights = np.squeeze(central_smooth),
                                         cmap = 'Greens',
                                         edgecolor = 'none',
                                         rasterized = True
                                         )
    
    cbar_fitted_smooth = plt.colorbar(fig_fitted_smooth[3], ax=axes[0,1], pad=0, label='Fit (finer binning)')
    
    cbar_fitted_smooth.ax.yaxis.label.set_size(cbar_fontsize)
    
    axes[0,1].set_xlabel('x0', fontsize = label_fontsize)
    axes[0,1].set_ylabel('x1', fontsize = label_fontsize)
    
    # Compute the candidate function and the error in original bins
    central_hist = func_evaluate(func_candidate = func_candidate,
                                 x = x,
                                 dim = 2,
                                 param_shifted = None,
                                 sigma_pm = None
                                 )
    
    central_error = y - central_hist
    
    if y_up is not None and y_down is not None:
        y_unc =  np.where(central_error < 0,
                          np.where(y_up != 0, y_up, y_down),
                          np.where(y_down != 0, y_down, y_up)
                          )
                          
        central_error = central_error / y_unc

    
    # Plot the candidate function in original bins
    fig_fitted_hist = axes[1,0].hist2d(x[:,0],
                                       x[:,1],
                                       bins = [x0_bins, x1_bins],
                                       weights = np.squeeze(central_hist),
                                       cmap = 'Greens',
                                       edgecolor = 'none',
                                       rasterized = True
                                       )
    
    cbar_fitted_hist = plt.colorbar(fig_fitted_hist[3], ax=axes[1,0], pad=0, label='Fit (same binning as Data)')
    
    cbar_fitted_hist.ax.yaxis.label.set_size(cbar_fontsize)
    
    axes[1,0].set_xlabel('x0', fontsize = label_fontsize)
    axes[1,0].set_ylabel('x1', fontsize = label_fontsize)

    # Plot the error in sigmas in original bins
    fig_error = axes[1,1].hist2d(x[:,0],
                                 x[:,1],
                                 bins = [x0_bins, x1_bins],
                                 weights = np.squeeze(central_error),
                                 cmap = 'bwr',
                                 vmin = -max(np.abs(central_error)),
                                 vmax = max(np.abs(central_error)),
                                 edgecolor = 'none',
                                 rasterized = True
                                 )
    
    if y_up is not None and y_down is not None:
        cbar_error = plt.colorbar(fig_error[3],
                                  ax = axes[1,1],
                                  pad = 0,
                                  label = '$\\frac{\\text{Data}-\\text{Fit}}{\\text{Data unc.}}$'
                                  )
        
    else:
        cbar_error = plt.colorbar(fig_error[3],
                                  ax = axes[1,1],
                                  pad = 0,
                                  label = '$\\text{Data}-\\text{Fit}$'
                                  )
                                  
    cbar_error.ax.yaxis.label.set_size(cbar_fontsize)
    
    axes[1,1].set_xlabel('x0', fontsize = label_fontsize)
    axes[1,1].set_ylabel('x1', fontsize = label_fontsize)
    
    
    # Define a string containing all the parameters in the function with each one in the form of
    # best-fit^{+1 sigma}_{-1 sigma}, with relative % error next to the +/-1 sigma values.
    # Highlight the one that is being shifted.
    parameters_string = ""
    
    parameters = func_candidate['Parameters: (best-fit, +1, -1)']
    
    for i in range(len(parameters)):
        # If the parameter is held fixed in the fit, it does not have +/-1 sigma values.
        if parameters[f'a{i+1}'][1] == 0:
            if f'a{i+1}' == param_shifted:
                parameters_string += r"$\bf{{{0} = {1}}}$".format(
                    f'a{i+1}',
                    round_a_number(parameters[f'a{i+1}'][0], 6),
                    )
                    
            else:
                parameters_string += r"${0} = {1}$".format(
                    f'a{i+1}',
                    round_a_number(parameters[f'a{i+1}'][0], 6),
                    )
                    
        else:
            if f'a{i+1}' == param_shifted:
                parameters_string += r"$\bf{{{0} = {1}^{{+ {2} ({4}\%) }}_{{- {3} ({5}\%)}}}}$".format(
                    f'a{i+1}',
                    round_a_number(parameters[f'a{i+1}'][0], 6),
                    round_a_number(parameters[f'a{i+1}'][1], 4),
                    round_a_number(np.abs(parameters[f'a{i+1}'][2]), 4),
                    round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][1]) / float(parameters[f'a{i+1}'][0])), 3),
                    round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][2]) / float(parameters[f'a{i+1}'][0])), 3)
                    )
                    
            else:
                parameters_string += r"$\text{{{0}}} = {1}^{{+ {2} ({4}\%) }}_{{- {3} ({5}\%)}}$".format(
                f'a{i+1}',
                round_a_number(parameters[f'a{i+1}'][0], 6),
                round_a_number(parameters[f'a{i+1}'][1], 4),
                round_a_number(np.abs(parameters[f'a{i+1}'][2]), 4),
                round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][1]) / float(parameters[f'a{i+1}'][0])), 3),
                round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][2]) / float(parameters[f'a{i+1}'][0])), 3)
                )
        
        if i < len(parameters) - 1:
            parameters_string += ",  "
            
        if np.mod(i + 1, 2) == 0:
            parameters_string += "\n"
            
    # Print parameter string defined above as title.
    title = textwrap.fill(func_candidate['Parameterized equation, unscaled'], width = 95) + "\n\n" + parameters_string
    
    axes[0,0].set_title(title, loc = 'left', size = 9.5)
    
    # Show the candidate # being plotted and some of the gof metrics.
    if y_up is not None and y_down is not None:
        axes[0,1].set_title(r"$\bfit{{Candidate\,\#{}}}$".format(candidate_idx) + "\n$\\chi^2$/NDF = {0}/{1}, p-value = {3}, RMSE = {2}".format(func_candidate['Chi2'], func_candidate['NDF'], func_candidate['RMSE'], round_a_number(func_candidate['p-value'], 4)), loc = 'right', size = 9.5)
        
    else:
        axes[0,1].set_title(r"$\bfit{{Candidate\,\#{}}}$".format(candidate_idx) + "\nRMSE = {0}, R2 = {1}".format(func_candidate['RMSE'], func_candidate['R2']), loc = 'right', size = 9.5)

    formatter = ticker.FormatStrFormatter('%.3g')

    
    # Add logo to every plot.
    logo_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'img/logo.png'
                             )
    logo_img = mpimg.imread(logo_path)
        
    bbox = axes[0,0].get_position()
    
    logo_width = 0.13
    logo_height = logo_width * (logo_img.shape[0] / logo_img.shape[1])

    ax_inset = fig.add_axes([bbox.x0 +0.73,
                            bbox.y1 + 0.085,
                            logo_width,
                            logo_height])

    ax_inset.imshow(logo_img)
    ax_inset.axis('off')
    
    plt.tight_layout()
    
    
def plot_all_syst_all_func_2D(
    func_candidates,
    x,
    bin_edges_2d,
    y,
    y_up,
    y_down,
    pdf_path,
    logy,
    logx
):
    '''
    Plot all candidate functions, each with all possible parameter variations.
    
    Arguments
    ---------
    func_candidates (pd.dataframe):
        Full dataframe containing all candidate functions after fits.
    
    x (np.ndarray):
        The independent variable.
        
    y (np.ndarray):
        The dependent variable.
    
    y_up (np.ndarray):
        +1 sigma of y.
        
    y_down (np.ndarray):
        -1 sigma of y.
        
    pdf_path (str):
        Save the output files to this directory.
    
    logy (bool):
        Plot y in log scale.
    
    logx (bool):
        Plot x in log scale.
    '''
    
    with PdfPages(pdf_path) as pdf:
        for i in range(len(func_candidates)):
            # Print candidate # page.
            fig, axes = plt.subplots(figsize = (9, 7))
            axes.axis('off')
            
            plt.text(0.5,
                     0.5,
                     'Candidate function #{}'.format(len(func_candidates) - 1 - i),
                     fontsize = 20,
                     ha = 'center',
                     va = 'center'
                     )
                     
            plt.tight_layout()
            
            plt.savefig(pdf, format = 'pdf')
            
            plt.close()
            
            if i < len(func_candidates) - 1:
                print('Plotting candidate functions {0}/{1} >>> {2}'.format(i+1, len(func_candidates), pdf_path), end='\r')
                
            else:
                print('Plotting candidate functions {0}/{1} >>> {2}'.format(i+1, len(func_candidates), pdf_path))
                
            func_candidate = func_candidates.iloc[len(func_candidates) - 1 - i]
            
            #if len(func_candidate['Parameters: (best-fit, +1, -1)']) > 0:
            if False: # fix later
                # Check if the function has any parameter that has variations.
                has_uncert = False
                for j in range(len(func_candidate['Parameters: (best-fit, +1, -1)'])):
                    if func_candidate['Parameters: (best-fit, +1, -1)'][f'a{j+1}'][1] > 0:
                        has_uncert = True
                
                # Plot central, up, and down in the same plot, otherwise plot only the central.
                if has_uncert:
                    for j in range(len(func_candidate['Parameters: (best-fit, +1, -1)'])):
                        if func_candidate['Parameters: (best-fit, +1, -1)'][f'a{j+1}'][1] > 0:
                            plot_single_syst_single_func_2D(func_candidate = func_candidate,
                                                            candidate_idx = len(func_candidates) - 1 - i,
                                                            x = x,
                                                            bin_edges_2d = bin_edges_2d,
                                                            y = y,
                                                            y_up = y_up,
                                                            y_down = y_down,
                                                            param_shifted = f'a{j+1}',
                                                            logy = logy,
                                                            logx = logx
                                                            )
                            
                            plt.savefig(pdf, format='pdf')
                            plt.close()
                else:
                    plot_single_syst_single_func_2D(func_candidate = func_candidate,
                                                    candidate_idx = len(func_candidates) - 1 - i,
                                                    x = x,
                                                    bin_edges_2d = bin_edges_2d,
                                                    y = y,
                                                    y_up = y_up,
                                                    y_down = y_down,
                                                    param_shifted = None,
                                                    logy = logy,
                                                    logx = logx
                                                    )
                    
                    plt.savefig(pdf, format='pdf')
                    plt.close()
                    
            else:
                plot_single_syst_single_func_2D(func_candidate = func_candidate,
                                                candidate_idx = len(func_candidates) - 1 - i,
                                                x = x,
                                                bin_edges_2d = bin_edges_2d,
                                                y = y,
                                                y_up = y_up,
                                                y_down = y_down,
                                                param_shifted = None,
                                                logy = logy,
                                                logx = logx
                                                )
                
                plt.savefig(pdf, format = 'pdf')
                plt.close()
    

def plot_correlation(
    func_candidate,
    candidate_idx,
    y_up,
    y_down
):
    '''
    Plot correlation matrix for the fitted parameters of a candidate function.
    
    Arguments
    ---------
    func_candidate (pd.dataframe):
        A particular candidate function (one row of the full func_candidates).
    
    candidate_idx (np.int):
        Candidate function # (ranked by function complexity).
    
    y_up (np.ndarray):
        +1 sigma of y.
        
    y_down (np.ndarray):
        -1 sigma of y.
    '''
    
    corr_dict = func_candidate['Correlation']
    
    # Retrieve the correlation for all pairs of fitted parameters.
    parameters = []
    
    pairs = list(corr_dict.keys())
    
    for pair in pairs:
        param1, param2 = pair.split(', ')
        
        parameters.append(param1)
        parameters.append(param2)
        
    parameters = sorted(list(set(parameters)), key = lambda x: (int(x[1:])))
    
    # Create the correlation matrix.
    df = pd.DataFrame(None, index = parameters, columns = parameters, dtype = float)
    
    for pair, value in corr_dict.items():
        param1, param2 = pair.split(', ')
        df.at[param2, param1] = value

    fig = plt.figure(figsize=(9, 8))
    
    sns.heatmap(df, annot = True, cmap = 'bwr', center = 0, vmin = -1, vmax = 1, fmt = '.3f')
    
    # Define the parameter string showing the best-fit and +/-1 sigma values.
    parameters_string = ""
    
    parameters = func_candidate['Parameters: (best-fit, +1, -1)']
    
    for i in range(len(parameters)):
        if parameters[f'a{i+1}'][1] == 0 and parameters[f'a{i+1}'][2] == 0:
            parameters_string += r"{0}$ = {1}$".format(
                f'a{i+1}',
                round_a_number(parameters[f'a{i+1}'][0], 6),
                )
                
        else:
            parameters_string += r"{0}$ = {1}^{{+ {2} ({4}\%) }}_{{- {3} ({5}\%)}}$".format(
                f'a{i+1}',
                round_a_number(parameters[f'a{i+1}'][0], 6),
                round_a_number(parameters[f'a{i+1}'][1], 4),
                round_a_number(np.abs(parameters[f'a{i+1}'][2]), 4),
                round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][1]) / float(parameters[f'a{i+1}'][0])), 3),
                round_a_number(100 * np.abs(float(parameters[f'a{i+1}'][2]) / float(parameters[f'a{i+1}'][0])), 3)
                )
                
        if i < len(parameters) - 1:
            parameters_string += ",  "
            
        if np.mod(i+1, 2) == 0:
            parameters_string += "\n"
    
    # Set the parameter string as the title.
    title = textwrap.fill(func_candidate['Parameterized equation, unscaled'], width=95) + "\n\n" + parameters_string
    
    plt.title(title, loc='left', size=9.5)
    
    # Display the candidate # being plotted and some of the gof metrics.
    if y_up is not None and y_down is not None:
        plt.title(r"$\bfit{{Candidate\,\#{}}}$".format(candidate_idx) + "\n$\\chi^2$/NDF = {0}/{1}, p-value = {3}, RMSE = {2}".format(func_candidate['Chi2'], func_candidate['NDF'], func_candidate['RMSE'], round_a_number(func_candidate['p-value'], 4)), loc='right', size=9.5)
        
    else:
        plt.title(r"$\bfit{{Candidate\,\#{}}}$".format(candidate_idx) + "\nRMSE = {0}, R2 = {1}".format(func_candidate['RMSE'], func_candidate['R2']), loc='right', size=9.5)
        
    plt.xticks(fontsize=10)
    plt.yticks(fontsize=10)
    
    # Add logo to every plot.
    logo_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'img/logo.png'
                             )
    logo_img = mpimg.imread(logo_path)
    
    logo_width = 0.1
    logo_height = logo_width * (logo_img.shape[0] / logo_img.shape[1])

    ax_inset = fig.add_axes([0.832,
                             0.955,
                             logo_width,
                             logo_height])

    ax_inset.imshow(logo_img)
    ax_inset.axis('off')
    
    plt.tight_layout()
    

def plot_all_corr(
    func_candidates,
    y_up,
    y_down,
    pdf_path
):
    '''
    Plot correlation matrix for all candidate functions at once.
    
    Arguments
    ---------
    func_candidates (pd.dataframe):
        Full dataframe containing all candidate functions after fits.
    
    y_up (np.ndarray):
        +1 sigma of y.
        
    y_down (np.ndarray):
        -1 sigma of y.
        
    pdf_path (str):
        Save the output files to this directory.
    '''
    
    with PdfPages(pdf_path) as pdf:
        for i in range(len(func_candidates)):
            # Print candidate # page.
            fig, axes = plt.subplots(figsize = (9, 7))
            axes.axis('off')
            
            plt.text(0.5,
                     0.5,
                     'Candidate function #{}'.format(len(func_candidates) - 1 - i),
                     fontsize = 20,
                     ha = 'center',
                     va = 'center'
                     )
                     
            plt.tight_layout()
            
            plt.savefig(pdf, format = 'pdf')
            
            plt.close()
        
            if i < len(func_candidates) - 1:
                print('Plotting correlation matrices {0}/{1} >>> {2}'.format(i+1, len(func_candidates), pdf_path), end='\r')
                
            else:
                print('Plotting correlation matrices {0}/{1} >>> {2}'.format(i+1, len(func_candidates), pdf_path))
                
            plot_correlation(func_candidate = func_candidates.iloc[len(func_candidates) - 1 - i],
                             candidate_idx = len(func_candidates) - 1 - i,
                             y_up = y_up,
                             y_down = y_down
                             )
            
            plt.savefig(pdf, format='pdf')
            plt.close()


def plot_gof(
    func_candidates,
    gof,
    log_scale = True
):
    '''
    Plot a goodness-of-fit metric vs. function complexity comparing all candidate functions.
    
    Arguments
    ---------
    func_candidates (pd.dataframe):
        Full dataframe containing all candidate functions after fits.
    
    gof (str):
        Which gof to plot.
        
    log_scale (bool):
        Plot gof in log scale.
    
    
    Returns
    -------
    Plot gof vs. function complexity.
    '''
    
    fig = plt.figure(figsize=(12,8))
    
    # Gof vs. complexity.
    plt.scatter(func_candidates['Complexity'], func_candidates[gof], marker = 'o', c = 'r', label = 'Candidate# indicated')
    
    # Display the candidate # next to the corresponding point.
    candidate_idx = list(range(len(func_candidates)))
    
    for i, idx in enumerate(candidate_idx):
        plt.annotate(idx,
                     (func_candidates['Complexity'][i], func_candidates[gof][i]),
                     textcoords = "offset points",
                     xytext = (0, 5),
                     ha = 'center',
                     size = 13
                     )
    
    plt.xticks(size = 18)
    plt.yticks(size = 18)
    
    plt.xlabel('Complexity', size = 20)
    plt.ylabel(gof, size = 20)
    
    plt.legend(loc = 'best', fontsize = 20)
    
    if log_scale:
        if gof != 'p-value':
            plt.yscale('log')
        
    plt.title(gof, loc='center', size=25)
    
    # Add logo to every plot.
    logo_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'img/logo.png'
                             )
    logo_img = mpimg.imread(logo_path)
    
    logo_width = 0.15
    logo_height = logo_width * (logo_img.shape[0] / logo_img.shape[1])
    
    ax_inset = fig.add_axes([0.85,
                             0.945,
                             logo_width,
                             logo_height])

    ax_inset.imshow(logo_img)
    ax_inset.axis('off')
    
    plt.tight_layout()
        
        
def plot_all_gof(
    func_candidates,
    y_up,
    y_down,
    pdf_path
):
    '''
    Plot all gof metrices at once.
    
    Arguments
    ---------
    func_candidates (pd.dataframe):
        Full dataframe containing all candidate functions after fits.
    
    y_up (np.ndarray):
        +1 sigma of y.
        
    y_down (np.ndarray):
        -1 sigma of y.
        
    pdf_path (str):
        Save the output files to this directory.
    
    
    Returns
    -------
    Plot all gof metrices at once and save to an output file.
    '''
    
    print('Plotting goodness-of-fit scores >>> {}'.format(pdf_path))
    
    # List of gof to plot.
    gof = ['RMSE', 'R2']
    if y_up is not None and y_down is not None:
        gof.append('Chi2/NDF')
        gof.append('p-value')
        
    with PdfPages(pdf_path) as pdf:
        for i in range(len(gof)):
            plot_gof(func_candidates = func_candidates,
                     gof = gof[i],
                     log_scale = True
                     )
            
            plt.savefig(pdf, format = 'pdf')
            
            plt.close()





from pysr import PySRRegressor
import sympy

pysr_config = PySRRegressor(
    model_selection = 'accuracy',
    niterations = 200,
    maxsize = 60,
    binary_operators = [
        '+', '*'
                     ],
    unary_operators = [
        'exp',
        'gauss(x) = exp(-x*x)',
        'tanh',
    ],
    nested_constraints = {
        'tanh':   {'tanh': 0, 'exp': 0, 'gauss': 0, '*': 2},
        'exp':    {'tanh': 0, 'exp': 0, 'gauss': 0, '*': 2},
        'gauss':  {'tanh': 0, 'exp': 0, 'gauss': 0, '*': 2},
        '*':      {'tanh': 1, 'exp': 1, 'gauss': 1, '*': 2},
    },
    extra_sympy_mappings={
        'gauss': lambda x: sympy.exp(-x*x),
                         },
    loss='loss(y, y_pred, weights) = (y - y_pred)^2 * weights',
)
